from PyQt5.QtWidgets import QPushButton
from PyQt5.QtGui import QIcon, QPixmap, QColor, QImage
from PyQt5.QtCore import QSize
import traceback
import numpy as np
from matplotlib.colors import LinearSegmentedColormap
import matplotlib.pyplot as plt

class PushButton(QPushButton):

    def __init__(self, element):
        super().__init__("", element.parent.widget)

        self.element = element

        self.clicked.connect(self.callback)

        if element.icon:
            self.setIcon(QIcon(element.icon))

        if element.colormap:
            self.update_colormap_icon()

        if element.text:
            if type(element.text) is str:
                txt = element.text
            else:
                txt = self.element.getvar(element.text.variable_group, element.text.variable)    
            self.setText(txt)    

        if self.element.tooltip:
            if type(self.element.tooltip) is str:
                txt = self.element.tooltip
            else:
                txt = self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable)    
            self.setToolTip(txt)

        self.set_geometry()

    def set(self):
        if type(self.element.text) is not str:
            self.setText(self.element.getvar(self.element.text.variable_group, self.element.text.variable))   
        if type(self.element.tooltip) is not str:
            self.setToolTip(self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable))
        if hasattr(self, "colormap"):
            self.update_colormap_icon()


    def callback(self):
        try:
            if self.element.callback and self.underMouse():
                self.element.callback(self)
                # Update GUI
                self.element.window.update()
        except:
            traceback.print_exc()

    def set_geometry(self):
        x0, y0, wdt, hgt = self.element.get_position()
        self.setGeometry(x0, y0, wdt, hgt)

    def update_colormap_icon(self):
        if type(self.element.colormap) is str:
            self.colormap = self.element.colormap
        else:
            self.colormap = self.element.getvar(self.element.colormap.variable_group, self.element.colormap.variable)
        cmap = plt.get_cmap(self.colormap) 
        x0, y0, wdt, hgt = self.element.get_position()
        # Create a QIcon from the colormap
        icon = self.create_icon_from_colormap(cmap, wdt, hgt)            
        # Set the icon on the button
        self.setIcon(icon)            
        # Set the icon size
        icon_size = QSize(wdt, hgt)  # Adjust the size as needed
        self.setGeometry(x0, y0, wdt, hgt)
        self.setIconSize(icon_size)

    def create_icon_from_colormap(self, cmap, wdt, hgt):
        # Create a gradient image
        gradient = np.linspace(0, 1.0, wdt)
        gradient = np.tile(gradient, (hgt, 1))
        # gradient = np.vstack((gradient, gradient))
        img = cmap(gradient)

        # Convert the image to a QPixmap
        img = np.uint8(img * 255)
        qimage = QImage(img.data, img.shape[1], img.shape[0], 
                        QImage.Format_RGBA8888)
        pixmap = QPixmap.fromImage(qimage)

        # Create a QIcon from the QPixmap
        return QIcon(pixmap)


from PySide6.QtWidgets import QButtonGroup, QRadioButton
import traceback

class RadioButtonGroup(QButtonGroup):

    def __init__(self, element):
        super().__init__(element.parent.widget)

        self.element = element

        if element.option_value.variable:
            group = element.option_value.variable_group
            name  = element.option_value.variable
            self.element.option_value.list = element.getvar(group, name)
        if element.option_string.variable:
            group = element.option_string.variable_group
            name  = element.option_string.variable
            self.element.option_string.list = element.getvar(group, name)
        for i in range(len(self.element.option_value.list)):
            d = QRadioButton(self.element.option_string.list[i], self.element.parent.widget)
            d.setVisible(True)
            d.id = i
            self.addButton(d)
            self.buttonClicked.connect(self.callback)

        self.set_geometry()

    def set(self):
        group  = self.element.variable_group
        name   = self.element.variable
        val    = self.element.getvar(group, name)
        values = self.element.option_value.list
        indx   = values.index(val)
        self.buttons()[indx].setChecked(True)

    def callback(self, button):
        group = self.element.variable_group
        name = self.element.variable
        values = self.element.option_value.list

        if self.element.getvar(group, name) == values[button.id]:
            # Value has not changed
            return

        self.element.setvar(group, name, values[button.id])

        try:
            if self.element.callback:
                self.element.callback(values[button.id], self)
                # Update GUI
                self.element.window.update()
        except:
            traceback.print_exc()

    def set_geometry(self):
        resize_factor = self.element.gui.resize_factor
        x0, y0, wdt, hgt = self.element.get_position()
        nbuttons = len(self.element.option_value.list)
        # Lower y of top button
        yll = y0 + hgt - int(nbuttons * 20 * resize_factor)
        for i in range(nbuttons):
            self.buttons()[i].setGeometry(x0,
                                          int(yll + i * 20 * resize_factor),
                                          wdt,
                                          int(20 * resize_factor))

    def set_enabled(self, true_or_false):
        nbuttons = len(self.element.option_value.list)
        for i in range(nbuttons):
            self.buttons()[i].setEnabled(true_or_false)

    def set_visible(self, true_or_false):
        nbuttons = len(self.element.option_value.list)
        for i in range(nbuttons):
            self.buttons()[i].setVisible(true_or_false)



from PyQt5.QtWidgets import QWidget, QTabWidget
import traceback

class TabPanel(QTabWidget):
    def __init__(self, element):
        super().__init__(element.parent.widget)

        self.element = element

        self.setVisible(True)

        for tab in element.tabs:
            # Place tab in tab panel
            widget = QWidget()
            tab.widget = widget
            self.addTab(widget, tab.text)

        # Add the callback
        self.currentChanged.connect(lambda indx, tabs=element.tabs: self.tab_selected(tabs, indx))

        self.previous_tab_index = 0

        self.set_geometry()

    def tab_selected(self, tabs, indx):

        # If the previous tab has a module, call the deselect method
        if tabs[self.previous_tab_index].module:
            if hasattr(tabs[self.previous_tab_index].module, "deselect"):
                try:
                    tabs[self.previous_tab_index].module.deselect()
                except:
                    traceback.print_exc()

        self.previous_tab_index = indx

        # Now call the select method
        if tabs[indx].module:
            if hasattr(tabs[indx].module, "select"):
                try:
                    tabs[indx].module.select()
                except:
                    traceback.print_exc()

        # Also check if there are tab panels in this tab. If so, execute select for active tab.
        if tabs[indx].elements:
            for element in tabs[indx].elements:
                if element.style == "tabpanel":
                    j = element.widget.currentIndex()
                    try:
                        element.widget.select_tab(j)
                    except:
                        traceback.print_exc()
        # Update GUI
        self.element.window.update()

    def select_tab(self, index):
        # Selected programmatically
        self.setCurrentIndex(index)
        self.tab_selected(self.element.tabs, index)

    def clear_tab(self, index):
        tab = self.element.tabs[index].widget
        for child in tab.children():
            child.setParent(None)

    def set_geometry(self):
        resize_factor = self.element.gui.resize_factor
        x0, y0, wdt, hgt = self.element.get_position()
        self.setGeometry(x0, y0, wdt, hgt)
        for tab in self.element.tabs:
            # Resize tab widgets
            tab.widget.setGeometry(0, 0, wdt, int(hgt - 20 * resize_factor))



from PySide6 import QtWebEngineWidgets, QtWebEngineCore, QtCore, QtWebChannel

import json
from geopandas import GeoDataFrame
from pandas import DataFrame
from pyproj import CRS, Transformer
import os
import sys
import requests

from guitares.map.layer import Layer, list_layers, find_layer_by_id

class WebEnginePage(QtWebEngineCore.QWebEnginePage):
    def __init__(self, view, print_messages):
        super().__init__(view)
        self.print_messages = print_messages

    def javaScriptConsoleMessage(self, level, message, lineNumber, sourceID):
        if self.print_messages:
            print(f"[JS] {message} (line {lineNumber}, source: {sourceID})")
            sys.stdout.flush()  # Ensures it appears even if buffering is active

    def javaScriptAlert(self, security_origin, msg):
        # Suppress alert()
        pass

    def javaScriptConfirm(self, security_origin, msg):
        # Suppress confirm(), return False by default
        return False

    def javaScriptPrompt(self, security_origin, msg, default):
        # Suppress prompt(), return empty string by default
        return ''


class MapLibre(QtCore.QObject):

    def __init__(self, element):
        super().__init__(element.parent.widget)

        self.gui = element.gui
        self.element = element
        self.nr_load_attempts = 0
        self.nr_ready_attempts = 0

        self.crs = CRS(4326)
        self.callback_module = element.module
        self.layer = {}
        self.map_extent = None
        self.map_center = None
        self.map_moved = None
        self.point_clicked_callback = None
        self.zoom = None

        self.url = "http://localhost:" + str(self.gui.server_port)

        # Check for internet connection
        try:
            requests.get("http://www.google.com", timeout=5)
        except requests.ConnectionError:
            print("No internet connection available.")
            map_style = "none"
            offline = True
        else:
            print("Internet connection available.")
            map_style = element.map_style
            offline = False

        # List all icon files in the icons folder
        icon_path = os.path.join(self.gui.server_path, "icons")
        icon_files = os.listdir(icon_path)
        icon_files = [f for f in icon_files if f.endswith(".png")]
        icon_list_string = ""
        for icon_file in icon_files:
            icon_list_string = icon_list_string + "'/icons/" + icon_file + "',"
        icon_list_string = "[" + icon_list_string + "]"    

        file_name = os.path.join(self.gui.server_path, "js", "defaults.js")
        with open(file_name, "w") as f:
            f.write("window.default_style = '" + map_style + "';\n")
            f.write("window.default_center = [" + str(element.map_center[0]) + "," + str(element.map_center[1]) + "]\n")
            f.write("window.default_zoom = " + str(element.map_zoom) + ";\n")
            f.write("window.default_projection = '" + element.map_projection + "';\n")
            f.write("window.iconUrls = " + icon_list_string + ";\n")
            if offline:
                f.write("window.offline = true;\n")
            else:
                f.write("window.offline = false;\n")

        self.webchannel_ok = False
        self.ready = False

        self.server_path = self.gui.server_path

        self.view = QtWebEngineWidgets.QWebEngineView(element.parent.widget)
        self.view.setPage(WebEnginePage(self.view, self.gui.js_messages))
        # self.view.page().profile().clearHttpCache()
        # self.view.page().settings().setAttribute(QtWebEngineCore.QWebEngineSettings.WebAttribute.WebGLEnabled, True)
        self.view.page().settings().setAttribute(QtWebEngineCore.QWebEngineSettings.LocalContentCanAccessRemoteUrls, True)

        self.channel = QtWebChannel.QWebChannel()
        self.channel.registerObject("MapLibre", self)
        self.view.page().setWebChannel(self.channel)

        self.set_geometry()
        self.view.loadFinished.connect(self.load_finished)

        print("Loading map ...")
        self.view.setUrl(QtCore.QUrl(self.url))

    def load_finished(self, message):
        self.timer_ping = QtCore.QTimer()        
        self.timer_ping.timeout.connect(self.ping)
        self.timer_ping.start(1000)

    def ping(self):
        # Sending a ping to main.js
        self.runjs("/js/main.js", "ping", arglist=["ping"])

    def set(self):
        pass

    def setVisible(self, visible):
        self.view.setVisible(visible)

    def set_geometry(self):
        x0, y0, wdt, hgt = self.element.get_position()
        self.view.setGeometry(x0, y0, wdt, hgt)

    def take_screenshot(self, output_file):
        self.view.grab().save(output_file, b"PNG")

    @QtCore.Slot(str)
    def pong(self, message):
        # Python heard a pong!
        self.timer_ping.stop()
        # Add map
        self.runjs("/js/main.js", "addMap", arglist=[])

    @QtCore.Slot(str)
    def mapReady(self, coords):
        coords = json.loads(coords)
        self.ready = True
        self.map_extent = coords
        if hasattr(self.callback_module, "map_ready"):
            self.callback_module.map_ready(self)
        # Set dependencies now
        self.element.set_dependencies()

    @QtCore.Slot(str)
    def layerStyleSet(self, coords):
        self.redraw_layers()

    @QtCore.Slot(str)
    def layerAdded(self, layer_id):
        layer = find_layer_by_id(layer_id, self.layer)
        layer.layer_added()

    @QtCore.Slot(str)
    def mouseMoved(self, coords):
        if hasattr(self.callback_module, "mouse_moved"):
            coords = json.loads(coords)
            lon = coords["lng"]
            lat = coords["lat"]
            if not self.crs.is_geographic:
                transformer = Transformer.from_crs(4326, self.crs, always_xy=True)
                x, y = transformer.transform(lon, lat)
            else:
                x = lon
                y = lat
            self.callback_module.mouse_moved(x, y, lon, lat)

    @QtCore.Slot(str)
    def mapMoved(self, coords):
        coords = json.loads(coords)
        self.map_extent = coords[0:2]
        self.map_center = coords[2:5]
        self.zoom = coords[4]
        # Loop through layers to update each
        layers = list_layers(self.layer)
        for layer in layers:
            layer.update()
        if hasattr(self.callback_module, "map_moved"):
            self.callback_module.map_moved(coords, self)

    @QtCore.Slot(str)
    def pointClicked(self, coords):
        coords = json.loads(coords)
        # Transform to local crs
        if self.crs.to_epsg() != 4326:
            transformer = Transformer.from_crs(4326, self.crs, always_xy=True)
            x, y = transformer.transform(coords["lng"], coords["lat"])
        else:
            x = coords["lng"]
            y = coords["lat"]
        if self.point_clicked_callback:
            self.point_clicked_callback(x, y)

    @QtCore.Slot(str)
    def getMapExtent(self, coords):
        coords = json.loads(coords)
        self.map_extent = coords

    @QtCore.Slot(str)
    def getMapCenter(self, coords):
        coords = json.loads(coords)
        self.map_center = coords

    @QtCore.Slot(str, str)
    def featureClicked(self, layer_id, feature_props):
        # Find layer by ID
        layer = find_layer_by_id(layer_id, self.layer)
        if hasattr(layer, "select"):
            if layer.select:
                layer.select(json.loads(feature_props), self)

    @QtCore.Slot(str, str, str)
    def featureDrawn(self, feature_collection, feature_id, layer_id):
        layer = find_layer_by_id(layer_id, self.layer)
        layer.feature_drawn(json.loads(feature_collection), feature_id)

    @QtCore.Slot(str, str, str)
    def featureModified(self, feature_collection, feature_id, layer_id):
        layer = find_layer_by_id(layer_id, self.layer)
        layer.feature_modified(json.loads(feature_collection), feature_id)

    @QtCore.Slot(str, str, str)
    def featureSelected(self, feature_collection, feature_id, layer_id):
        layer = find_layer_by_id(layer_id, self.layer)
        layer.feature_selected(json.loads(feature_collection), feature_id)

    @QtCore.Slot(str)
    def featureDeselected(self, layer_id):
        layer = find_layer_by_id(layer_id, self.layer)
        if layer:
            layer.feature_deselected()

    @QtCore.Slot(str, str, str)
    def featureAdded(self, feature_collection, feature_id, layer_id):
        layer = find_layer_by_id(layer_id, self.layer)
        layer.feature_added(json.loads(feature_collection), feature_id)

    def get_extent(self):
        js_string = "import('/js/main.js').then(module => {module.getExtent()});"
        self.view.page().runJavaScript(js_string)

    def get_center(self):
        js_string = "import('/js/main.js').then(module => {module.getCenter()});"
        self.view.page().runJavaScript(js_string)

    def click_point(self, callback):
        self.point_clicked_callback = callback
        self.runjs("/js/main.js", "clickPoint")

    def set_center(self, lon, lat):
        self.runjs("/js/main.js", "setCenter", arglist=[lon, lat])

    def set_zoom(self, zoom):
        self.runjs("/js/main.js", "setZoom", arglist=[zoom])

    def fit_bounds(self, lon1, lat1, lon2, lat2, crs=None):
        if crs is not None:
            if not crs.is_geographic:
                # Convert to lat/lon
                transformer = Transformer.from_crs(crs, 4326, always_xy=True)
                lon1, lat1 = transformer.transform(lon1, lat1)
                lon2, lat2 = transformer.transform(lon2, lat2)
        self.runjs("/js/main.js", "fitBounds", arglist=[lon1, lat1, lon2, lat2])

    def jump_to(self, lon, lat, zoom):
        self.runjs("/js/main.js", "jumpTo", arglist=[lon, lat, zoom])

    def fly_to(self, lon, lat, zoom):
        self.runjs("/js/main.js", "flyTo", arglist=[lon, lat, zoom])

    def set_projection(self, projection):
        self.runjs("/js/main.js", "setProjection", arglist=[projection])

    def set_layer_style(self, style):
        self.runjs("/js/main.js", "setLayerStyle", arglist=[style])

    def set_terrain(self, true_or_false, exaggeration):
        self.runjs("/js/main.js", "setTerrain", arglist=[true_or_false, exaggeration])

    def set_mouse_default(self):
        self.runjs("/js/main.js", "setMouseDefault", arglist=[])
        self.runjs("/js/draw_layer.js", "setMouseDefault", arglist=[])

    def close_popup(self):
        self.runjs("/js/main.js", "closePopup")

    def add_layer(self, layer_id):
        # Adds a container layer
        if layer_id not in self.layer:
            self.layer[layer_id] = Layer(self, layer_id, layer_id)
            self.layer[layer_id].map_id = layer_id
        else:
            print("Layer " + layer_id + " already exists.")
        return self.layer[layer_id]

    def list_layers(self):
        # Return a list with all layers
        return list_layers(self.layer)

    def redraw_layers(self):
        # Redraw all layers (after map style has changed)
        # First clear the layer list in the draw_layer.js file
        self.runjs(
            "/js/draw_layer.js",
            "clearLayerList"
        )
        layers = self.list_layers()
        for layer in layers:
            layer.redraw()

    def compare(self):
        self.runjs("/js/main.js", "compare", arglist=[])

    def runjs(self, module, function, arglist=None):
        if not arglist:
            arglist = []
        string = "import('" + self.url + module + "').then(module => {module." + function + "("
        # string = "import(" + module + "').then(module => {module." + function + "("
        for iarg, arg in enumerate(arglist):
            if isinstance(arg, bool):
                if arg:
                    string = string + "true"
                else:
                    string = string + "false"
            elif isinstance(arg, int):
                string = string + str(arg)
            elif isinstance(arg, float):
                string = string + str(arg)
            elif isinstance(arg, dict):
                string = string + json.dumps(arg).replace('"',"'")
            elif isinstance(arg, list):
                string = string + json.dumps(arg).replace('"',"'")
            elif isinstance(arg, tuple):
                string = string + json.dumps(arg).replace('"',"'")
            elif isinstance(arg, GeoDataFrame):
                if len(arg) == 0:
                    string = string + "{}"
                else:
                    # Need to remove timeseries from geodataframe
                    for columnName, columnData in arg.items():
                        if isinstance(columnData.iloc[0], DataFrame):
                            arg = arg.drop([columnName], axis=1)
                    string = string + arg.to_json()
            elif arg is None:
                string = string + "null"        
            else:
                string = string + "'" + arg + "'"
            if iarg < len(arglist) - 1:
                string = string + ","
        string = string + ")});"
        # print(string)
        self.view.page().runJavaScript(string)



import importlib

from guitares.dependencies import Dependency, DependencyCheck

class Position:
    def __init__(self):
        self.x      = 0
        self.y      = 0
        self.width  = 10
        self.height = 10

class Text:
    def __init__(self, variable_group):
        self.variable       = ""
        self.variable_group = variable_group

class ColorMap:
    def __init__(self, variable_group):
        self.variable       = ""
        self.variable_group = variable_group

class OptionValue:
    def __init__(self, variable_group):
        self.list           = []
        self.variable       = ""
        self.variable_group = variable_group

class OptionString:
    def __init__(self, variable_group):
        self.list           = []
        self.variable       = ""
        self.variable_group = variable_group

class Tab:
    def __init__(self, variable_group, module):
        self.text           = ""
        self.module         = ""
        self.variable_group = variable_group
        self.module         = module
        self.elements       = []
        self.gui            = None

class Element:
    def __init__(self, dct, parent, window):
        self.style = None
        self.widget = None
        self.parent = parent
        self.position = Position()
        self.tabs = []
        self.dependencies = []
        self.elements = []
        self.visible = True
        self.enable = True
        self.id = ""
        self.variable = None
        self.variable_group = parent.variable_group
        self.module = parent.module
        self.method = ""
        self.callback = None
        self.text = ""
        self.text_position = "left"
        self.tooltip = ""
        self.type = str
        self.gui    = window.gui
        self.getvar = window.gui.getvar
        self.setvar = window.gui.setvar
        self.window = window
        self.select = "item"
        self.option_value = OptionValue(self.variable_group)
        self.option_string = OptionString(self.variable_group)
        self.title = "Select File"
        self.filter = "All Files (*.*)"
        self.url = ""
        self.icon = ""
        self.collapse = False
        self.collapsed = False
        self.fraction_collapsed = 1.0
        self.fraction_expanded = 0.5
        self.multiselection = False
        self.sortable = True
        self.selection_type = "single"
        self.colormap = None
        self.ready = False
        self.full_path = False
        self.allow_directory_change = True

        # Map
        if self.gui.map_engine == "mapbox":
            self.map_style = "mapbox://styles/mapbox/streets-v12"
        elif self.gui.map_engine == "maplibre":
            self.map_style = "osm"
        else:
            # This should not happen
            self.map_style = "osm"

        self.map_center = [0, 0]
        self.map_zoom = 0
        self.map_projection = "mercator"

        # Now update element attributes based on dict

        self.style = dct["style"]

        if "position" in dct:
            self.position.x = dct["position"]["x"]
            self.position.y = dct["position"]["y"]
            if "width" in dct["position"]:
                self.position.width = dct["position"]["width"]
            if "height" in dct["position"]:
                self.position.height = dct["position"]["height"]

        if "variable" in dct:
            self.variable = dct["variable"]

        if "variable_group" in dct:
            self.variable_group = dct["variable_group"]

        if "id" in dct:
            self.id = dct["id"]

        if "module" in dct:
            if isinstance(dct["module"], str):
                try:
                    self.module = importlib.import_module(dct["module"])
                except:
                    print("Error! Could not import module " + dct["module"])

        if "method" in dct:
            self.method = dct["method"]

        if hasattr(self.method, '__call__'):
             # Callback function is already defined as method
             self.callback = self.method
        else:
            if self.module and self.method:
                try:
                    # Start with the base module
                    module = self.module
                    # If the method contains a dot, it means that it is a method of a class
                    method_path = self.method.split(".")
                    # Loop through the method path
                    for idx, method in enumerate(method_path):
                        # The last element is the method itself
                        if idx == len(method_path) - 1:
                            if hasattr(module, method):
                                self.callback = getattr(module, method)
                            else:
                                raise Exception("Error! Could not find method " + self.method + " in module " + self.module.__name__)
                        else:
                            if hasattr(module, method):
                                class_ = getattr(module, method)
                                # Initialize the class object
                                module = class_()
                            else:
                                raise Exception("Error! Could not find method " + self.method + " in module " + self.module.__name__)
                except Exception as e:
                    print(e)
        if self.variable_group in self.gui.variables:
            if self.variable in self.gui.variables[self.variable_group]:
                self.type = type(self.gui.variables[self.variable_group][self.variable]["value"])

        # "title" for backward compatibility
        if "title" in dct:
            self.text = dct["title"]
            self.title = dct["title"]

        if "text" in dct:
            if isinstance(dct["text"], dict):
                self.text = Text(self.variable_group)
                if "variable" in dct["text"]:
                    self.text.variable = dct["text"]["variable"]
                if "variable_group" in dct["text"]:
                    self.text.variable_group = dct["text"]["variable_group"]
            else:
                self.text = dct["text"]

        if "textposition" in dct:
            self.text_position = dct["textposition"]

        if "text-position" in dct:
            self.text_position = dct["text-position"]

        if "text_position" in dct:
            self.text_position = dct["text_position"]

        if "colormap" in dct:
            if isinstance(dct["colormap"], dict):
                self.colormap = ColorMap(self.variable_group)
                if "variable" in dct["colormap"]:
                    self.colormap.variable = dct["colormap"]["variable"]
                if "variable_group" in dct["colormap"]:
                    self.colormap.variable_group = dct["colormap"]["variable_group"]
            else:
                self.colormap = dct["colormap"]

        # For backward compatibility
        if "tooltipstring" in dct:
            dct["tooltip"] = dct["tooltipstring"]

        if "tooltip" in dct:    
            if isinstance(dct["tooltip"], dict):
                self.tooltip = Text(self.variable_group)
                if "variable" in dct["tooltip"]:
                    self.tooltip.variable = dct["tooltip"]["variable"]
                if "variable_group" in dct["tooltip"]:
                    self.tooltip.variable_group = dct["tooltip"]["variable_group"]
            else:
                self.tooltip = dct["tooltip"]

        # Special for popupmenus and listboxes
        if "select" in dct:
            # Either 'index' or 'item'
            self.select = dct["select"]

        if "option_value" in dct:    
            if isinstance(dct["option_value"], dict):
                if "variable" in dct["option_value"]:
                    self.option_value.variable = dct["option_value"]["variable"]
                if "variable_group" in dct["option_value"]:
                    self.option_value.variable_group = dct["option_value"]["variable_group"]
            else:
                # It's a list
                self.option_value.list = dct["option_value"]
                # Loop through the list and convert to the correct type
                for idx, item in enumerate(self.option_value.list):
                    # Check if item is a string
                    if isinstance(item, str):
                        if item == "true":
                            self.option_value.list[idx] = True
                        elif item == "false":
                            self.option_value.list[idx] = False   

        if "option_string" in dct:    
            if isinstance(dct["option_string"], dict):
                if "variable" in dct["option_string"]:
                    self.option_string.variable = dct["option_string"]["variable"]
                if "variable_group" in dct["option_string"]:
                    self.option_string.variable_group = dct["option_string"]["variable_group"]
            else:
                # It's a list
                self.option_string.list = dct["option_string"]

        if "filter" in dct:
            if isinstance(dct["filter"], dict):
                self.filter = Text(self.variable_group)
                if "variable" in dct["filter"]:
                    self.filter.variable = dct["filter"]["variable"]
                if "variable_group" in dct["filter"]:
                    self.filter.variable_group = dct["filter"]["variable_group"]
            else:
                self.filter = dct["filter"]

        if "full_path" in dct:
            self.full_path = dct["full_path"]

        if "allow_directory_change" in dct:
            self.allow_directory_change = dct["allow_directory_change"]

        if "url" in dct:
            if isinstance(dct["url"], dict):
                self.url = Text(self.variable_group)
                if "variable" in dct["url"]:
                    self.url.variable = dct["url"]["variable"]
                if "variable_group" in dct["url"]:
                    self.url.variable_group = dct["url"]["variable_group"]
            else:
                self.url = dct["url"]

        if "collapse" in dct:
            self.collapse = dct["collapse"]

        if "collapsed" in dct:
            self.collapsed = dct["collapsed"]

        if "fraction_collapsed" in dct:
            self.fraction_collapsed = dct["fraction_collapsed"]

        if "fraction_expanded" in dct:
            self.fraction_expanded = dct["fraction_expanded"]

        if "multiselection" in dct:
            self.multiselection = dct["multiselection"]

        if "enable" in dct:
            self.enable = dct["enable"]

        if "selection_type" in dct:
            self.selection_type = dct["selection_type"]

        if "selection_direction" in dct:
            self.selection_direction = dct["selection_direction"]

        if "sortable" in dct:
            self.sortable = dct["sortable"]

        if "map_style" in dct:
            self.map_style = dct["map_style"]

        if "map_lat" in dct:
            self.map_center[1] = dct["map_lat"]

        if "map_lon" in dct:
            self.map_center[0] = dct["map_lon"]

        if "map_zoom" in dct:
            self.map_zoom = dct["map_zoom"]

        if "map_projection" in dct:
            self.map_projection = dct["map_projection"]    

        if "dependency" in dct:
            for dep in dct["dependency"]:
                dependency = Dependency()
                dependency.gui = parent.gui
                if "action" in dep:
                    dependency.action = dep["action"]
                if "checkfor" in dep:
                    dependency.checkfor = dep["checkfor"]
                for check_dct in dep["check"]:
                    check = DependencyCheck(self.variable_group)
                    if "variable" in check_dct:
                        check.variable = check_dct["variable"]
                    if "variable_group" in check_dct:
                        check.variable_group = check_dct["variable_group"]
                    if "operator" in check_dct:
                        check.operator = check_dct["operator"]
                    if "value" in check_dct:
                        check.value = check_dct["value"]
                    dependency.checks.append(check)
                self.dependencies.append(dependency)

        if self.style == "tabpanel":

            self.tabs = []

            # Loop through tabs
            for itab, tab_dct in enumerate(dct["tab"]):

                tab = Tab(self.variable_group, self.module)
                tab.gui = self.gui
                
                if "string" in tab_dct: # Backward compatibility
                    tab.text = tab_dct["string"]
                if "text" in tab_dct:
                    tab.text = tab_dct["text"]
                if "variable_group" in tab_dct:
                    tab.variable_group = tab_dct["variable_group"]
                if "module" in tab_dct:
                    try:
                        if tab_dct["module"]:
                            tab.module = importlib.import_module(tab_dct["module"])
                    except Exception as e:
                        print("Error! Module " + tab_dct["module"] + " could not be imported!")
                        print(e)

                tab.dependencies = []
                if "dependency" in tab_dct:
                    for dep in tab_dct["dependency"]:
                        dependency = Dependency()
                        dependency.gui = self.gui
                        if "action" in dep:
                            dependency.action = dep["action"]
                        if "checkfor" in dep:
                            dependency.checkfor = dep["checkfor"]
                        for check_dct in dep["check"]:
                            check = DependencyCheck(self.variable_group)
                            if "variable" in check_dct:
                                check.variable = check_dct["variable"]
                            if "variable_group" in check_dct:
                                check.variable_group = check_dct["variable_group"]
                            if "operator" in check_dct:
                                check.operator = check_dct["operator"]
                            if "value" in check_dct:
                                check.value = check_dct["value"]
                            dependency.checks.append(check)
                        tab.dependencies.append(dependency)

                self.tabs.append(tab)


    def add(self):
        
        if self.style == "tabpanel":
            mod = importlib.import_module(f"guitares.{self.gui.framework}.tabpanel")
            self.widget = mod.TabPanel(self)

        elif self.style == "panel":
            # Add frame
            mod = importlib.import_module(f"guitares.{self.gui.framework}.frame")
            self.widget = mod.Frame(self)

        elif self.style == "dual_frame":
            # Add dual frame
            mod = importlib.import_module(f"guitares.{self.gui.framework}.dual_frame")
            self.widget = mod.DualFrame(self)

        elif self.style == "pushbutton":
            mod = importlib.import_module(f"guitares.{self.gui.framework}.pushbutton")
            self.widget = mod.PushButton(self)

        elif self.style == "edit":
            mod = importlib.import_module(f"guitares.{self.gui.framework}.edit")
            self.widget = mod.Edit(self)

        elif self.style == "datetimeedit":
            mod = importlib.import_module(f"guitares.{self.gui.framework}.date_edit")
            self.widget = mod.DateEdit(self)

        elif self.style == "text":
            mod = importlib.import_module(f"guitares.{self.gui.framework}.text")
            self.widget = mod.Text(self)

        elif self.style == "popupmenu":
            mod = importlib.import_module(f"guitares.{self.gui.framework}.popupmenu")
            self.widget = mod.PopupMenu(self)

        elif self.style == "listbox":
            mod = importlib.import_module(f"guitares.{self.gui.framework}.listbox")
            self.widget = mod.ListBox(self)

        elif self.style == "tableview":
            mod = importlib.import_module(f"guitares.{self.gui.framework}.tableview")
            self.widget = mod.TableView(self)

        elif self.style == "checkbox":
            mod = importlib.import_module(f"guitares.{self.gui.framework}.checkbox")
            self.widget = mod.CheckBox(self)

        elif self.style == "radiobuttongroup":
            mod = importlib.import_module(f"guitares.{self.gui.framework}.radiobuttongroup")
            self.widget = mod.RadioButtonGroup(self)

        elif self.style == "slider":
            mod = importlib.import_module(f"guitares.{self.gui.framework}.slider")
            self.widget = mod.Slider(self)

        elif self.style == "pushselectfile":
            mod = importlib.import_module(f"guitares.{self.gui.framework}.pushopenfile")
            self.widget = mod.PushOpenFile(self)

        elif self.style == "pushselectdir":
            mod = importlib.import_module(f"guitares.{self.gui.framework}.pushopendir")
            self.widget = mod.PushOpenDir(self)
            
        elif self.style == "pushsavefile":
            mod = importlib.import_module(f"guitares.{self.gui.framework}.pushsavefile")
            self.widget = mod.PushSaveFile(self)

        elif self.style == "map" or self.style == "mapbox" or self.style == "maplibre":

            okay = True
            if self.dependencies:
                for dep in self.dependencies:
                    if dep.action == "visible":
                        if not dep.get():
                            okay = False
            if okay:                
                # Determine which map to use
                if self.gui.map_engine == "mapbox":
                    mod = importlib.import_module(f"guitares.{self.gui.framework}.mapbox")
                    self.widget = mod.MapBox(self)
                elif self.gui.map_engine == "maplibre":
                    mod = importlib.import_module(f"guitares.{self.gui.framework}.maplibre")
                    self.widget = mod.MapLibre(self)

        elif self.style == "mapbox_compare" or self.style == "maplibre_compare" or self.style == "map_compare":
            # We don't want to add the mapbox widget if set to invisible, because it takes a long time to load.
            # This means that widgets that are originally set to invisible will not be added to the GUI!
            okay = True
            if self.dependencies:
                for dep in self.dependencies:
                    if dep.action == "visible":
                        if not dep.get():
                            okay = False
            if okay:
                if self.gui.map_engine == "mapbox":
                    mod = importlib.import_module(f"guitares.{self.gui.framework}.mapbox_compare")
                    self.widget = mod.MapBoxCompare(self)
                elif self.gui.map_engine == "maplibre":
                    mod = importlib.import_module(f"guitares.{self.gui.framework}.maplibre_compare")
                    self.widget = mod.MapLibreCompare(self)

        elif self.style == "webpage":
            mod = importlib.import_module(f"guitares.{self.gui.framework}.webpage")
            self.widget = mod.WebPage(self)

        else:
            print("Element style " + self.style + " not recognized!")

        # And set the visibility
        if self.style != "radiobuttongroup": # Cannot set radiobutton group to visible
            if self.widget:
                self.widget.setVisible(True)
        else:
            self.widget.set_visible(True)        

    def set_geometry(self):
        try:
            self.widget.set_geometry()
        except:
            pass    

    def get_position(self):

        # Position as defined in the yml file 
        position = self.position
        resize_factor = self.gui.resize_factor
        parent = self.parent.widget

        # TO DO: relative positions!
        pwdt = parent.geometry().width()
        phgt = parent.geometry().height()

        # Multiply with resize factor for high-res screens
        x0 = position.x * resize_factor
        y0 = position.y * resize_factor
        wdt = position.width * resize_factor
        hgt = position.height * resize_factor

        if x0>0:
            # x0 is absolute
            if wdt>0:
                # wdt is absolute
                pass
            else:
                # wdt is relative
                wdt = pwdt - x0 + wdt
        else:
            # x0 is relative
            if wdt>0:
                # wdt is absolute
                x0 = pwdt - wdt + x0
            else:
                # wdt is relative
                x0 = pwdt + x0
                wdt = pwdt - x0 + wdt

        if y0>0:
            # y0 is absolute
            if hgt>0:
                # hgt is absolute
                y0 = phgt - (y0 + hgt)
            else:
                # hgt is relative
                y0 = - hgt
                hgt = phgt - position.y * resize_factor + hgt
        else:
            # y0 is relative
            if hgt>0:
                # hgt is absolute
                y0 = phgt - hgt
            else:
                # hgt is relative
                hgt = - y0 - hgt
                y0 = phgt - (y0 + hgt)

        # Round to integers
        x0 = int(x0)
        y0 = int(y0)
        wdt = int(wdt)
        hgt = int(hgt)

        return x0, y0, wdt, hgt

    def set_dependencies(self):
        # TODO: Would be cleaner to add enable and disable methods in the individual elements
        # Check if this element is a tabpanel
        if self.style == "tabpanel":
            for itab, tab in enumerate(self.tabs):
                # Check if this tab has dependencies
                if tab.dependencies:
                    for dependency in tab.dependencies:
                        true_or_false = dependency.get()
                        if dependency.action == "enable":
                            if true_or_false:
                                self.widget.setTabEnabled(itab, True)
                            else:
                                self.widget.setTabEnabled(itab, False)

                        elif dependency.action == "visible":
                            if true_or_false:
                                self.widget.setTabVisible(itab, True)
                            else:
                                self.widget.setTabVisible(itab, False)    

        else:            
            # "normal" element
            for dependency in self.dependencies:
                true_or_false = dependency.get()
                if dependency.action == "visible":
                    if self.style == "radiobuttongroup": # Cannot set radiobutton group directly
                        self.widget.set_visible(true_or_false)        
                    elif self.style == "mapbox" or self.style == "mapbox_compare" or self.style == "maplibre" or self.style == "maplibre_compare" or self.style == "map" or self.style == "map_compare":
                        if self.widget.ready:
                            self.widget.view.setVisible(true_or_false)        
                    else:    
                        if self.widget:
                            self.widget.setVisible(true_or_false)
                            if hasattr(self.widget, "text_widget"):
                                self.widget.text_widget.setVisible(true_or_false)
                elif dependency.action == "enable":
                    if self.style == "radiobuttongroup": # Cannot set radiobutton group directly
                        self.widget.set_enabled(true_or_false)        
                    else:    
                        if self.widget:
                            self.widget.setEnabled(true_or_false)
                            if hasattr(self.widget, "text_widget"):
                                self.widget.text_widget.setEnabled(true_or_false)
                elif dependency.action == "check":
                    self.widget.setChecked(true_or_false)
            if not self.enable:        
                if self.widget:
                    self.widget.setEnabled(False)
                    if hasattr(self.widget, "text_widget"):
                        self.widget.text_widget.setEnabled(False)

    def clear_tab(self, index):
        self.widget.clear_tab(index)

    def set_collapsed(self, true_or_false):
        self.collapsed = true_or_false
        self.gui.window.resize_elements(self.elements)



from PySide6.QtWidgets import (
    QMessageBox,
    QDialogButtonBox,
    QVBoxLayout,
    QLabel,
    QDialog,
    QProgressDialog,
    QFileDialog,
    QLineEdit,
    QApplication
)
from PySide6 import QtCore
import time
import os

class CustomDialog(QDialog):
    def __init__(self, text, title, button_text):
        super().__init__()
        self.setWindowTitle(title)
        self.buttonBox = QDialogButtonBox()
        for txt in button_text:
            self.buttonBox.addButton("  " + txt + "  ", QDialogButtonBox.AcceptRole)
        self.buttonBox.clicked.connect(self.clicked)
        self.buttonBox.accepted.connect(self.accept)
        self.layout = QVBoxLayout()
        message = QLabel(text)
        self.layout.addWidget(message)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

    def clicked(self, *args):
        self.answer = args[0].text().strip()

class StringDialog(QDialog):
    def __init__(self, text, title, string):
        super().__init__()
        self.setWindowTitle(title)
        self.buttonBox = QDialogButtonBox()
        self.buttonBox.addButton("  " + "OK" + "  ", QDialogButtonBox.AcceptRole)
        self.buttonBox.addButton("  " + "Cancel" + "  ", QDialogButtonBox.AcceptRole)
        self.buttonBox.clicked.connect(self.clicked)
        self.buttonBox.accepted.connect(self.accept)
        self.layout = QVBoxLayout()
        message = QLabel(text)
        self.layout.addWidget(message)
        self.edit = QLineEdit(string)
        self.layout.addWidget(self.edit)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

    def clicked(self, *args):
        answer = args[0].text().strip()
        if answer == "OK":
            self.answer = (self.edit.text(), True)
        else:
            self.answer = ("", False)

class ProgressDialog(QProgressDialog):
    def __init__(self, window, text, title, nmax):
        super().__init__(text, "Abort", 0, nmax, window)
        self.setWindowModality(QtCore.Qt.WindowModal)
        # self.setMinimum(1)
        # self.setMaximum(nmax)
        self.setMinimum(0)
        self.setMaximum(100)
        self.setWindowTitle(title)
        self.time_elapsed = QLabel("Time elapsed :", self)
        self.time_elapsed.setGeometry(15, 65, 200, 20)
        self.time_remaining = QLabel("Est. time remaining :", self)
        self.time_remaining.setGeometry(15, 85, 200, 20)
        self.t0 = time.time()
        self.nmax = nmax
        self.setMinimumDuration(100)
        self.setValue(0)
        self.show()

    def set_value(self, i):
        self.setValue(i)
        t = time.time()
        dt = t - self.t0
        if dt > 60.0:
            tmin = int(dt / 60)
            tsec = dt - tmin * 60.0
            new_string = (
                "Time elapsed : " + str(int(tmin)) + "m " + str(int(tsec)) + "s"
            )
        else:
            new_string = "Time elapsed : " + str(int(dt)) + "s"

        self.time_elapsed.setText(new_string)
        frac = max(i, 1) / self.nmax
        trem = dt / frac - dt

        if trem > 60.0:
            tmin = int(trem / 60)
            tsec = trem - tmin * 60.0
            new_string = (
                "Est. time remaining : " + str(int(tmin)) + "m " + str(int(tsec)) + "s"
            )
        else:
            new_string = "Est. time remaining : " + str(int(trem)) + "s"
        self.time_remaining.setText(new_string)
        self.show()
        QApplication.processEvents()

    def set_text(self, text):
        self.setLabelText(text)

    def set_minimum(self, nmin):
        self.nmax = nmin
        self.setMaximum(nmin)

    def set_maximum(self, nmax):
        self.nmax = nmax
        self.setMaximum(nmax)

    def was_canceled(self):
        QApplication.processEvents()
        return self.wasCanceled()


class WaitDialog(QProgressDialog):
    def __init__(self, window, text, title):
        super().__init__(text, "Abort", 0, 100, window)
        self.setWindowModality(QtCore.Qt.WindowModal)
        self.setMinimumDuration(1)
        self.setMaximum(100)
        self.setWindowTitle(title)
        self.children()[1].setVisible(False)  # Make progress bar invisible
        self.children()[3].setVisible(False)  # Make push button invisible
        self.show()

    def set_value(self, i):
        self.setValue(i)

# class StringDialog(QProgressDialog):
#     def __init__(self, window, text, title):
#         super().__init__(text, "Abort", 0, 100, window)
#         self.setWindowModality(QtCore.Qt.WindowModal)
#         self.setMinimumDuration(1)
#         self.setMaximum(100)
#         self.setWindowTitle(title)
#         self.children()[1].setVisible(False)  # Make progress bar invisible
#         self.children()[3].setVisible(False)  # Make push button invisible
#         self.show()

#     def set_value(self, i):
#         self.setValue(i)

def dialog(
    window,
    text,
    title=" ",
    type="warning",
    button_text=None,
    nmax=100,
    cancel=None,
    filter=None,
    selected_filter=None,
    path=None,
    file_name=None,
    string=""
):
    if type == "warning":
        ret = QMessageBox.warning(window, title, text, QMessageBox.Ok, QMessageBox.Ok)
        return ret
    elif type == "info":
        ret = QMessageBox.information(
            window, title, text, QMessageBox.Ok, QMessageBox.Ok
        )
        return ret
    elif type == "critical":
        ret = QMessageBox.critical(window, title, text, QMessageBox.Ok, QMessageBox.Ok)
        return ret
    elif type == "question":
        ret = QMessageBox.question(
            window, title, text, QMessageBox.Ok | QMessageBox.Cancel, QMessageBox.Cancel
        )
        if ret == QMessageBox.Ok:
            return True
        else:
            return False
    elif type == "question_yes_no":
        ret = QMessageBox.question(
            window, title, text, QMessageBox.Yes | QMessageBox.No, QMessageBox.No
        )
        if ret == QMessageBox.Yes:
            return True
        else:
            return False
    elif type == "custom":
        dlg = CustomDialog(text, title, button_text)
        dlg.exec()
        return dlg.answer
    elif type == "progress":
        dlg = ProgressDialog(window, text, title, nmax)
        return dlg
    elif type == "wait":
        dlg = WaitDialog(window, text, title)
        for i in range(5):
            time.sleep(0.001)
            dlg.set_value(i)
        return dlg
    elif type == "open_file":
        if path is None:
            path = os.getcwd()
        if file_name:
            path = os.path.join(path, file_name)
        fname = QFileDialog.getOpenFileName(window, text, path, filter, selected_filter)
        return fname
    elif type == "save_file":
        if path is None:
            path = os.getcwd()
        if file_name:
            path = os.path.join(path, file_name)
        fname = QFileDialog.getSaveFileName(window, text, path, filter, selected_filter)
        return fname
    elif type == "select_path":
        if path is None:
            path = os.getcwd()
        new_path = QFileDialog.getExistingDirectory(window, text, path)
        return new_path
    elif type == "string":
        dlg = StringDialog(text, title, string)
        dlg.exec()
        return dlg.answer[0], dlg.answer[1]



import os
import glob
import numpy as np
from .colorbar import ColorBar
from .layer import Layer
from guitares.colormap import cm2png

from cht_tiling.tiling import make_floodmap_overlay_v2, make_topo_overlay_v2
from cht_tiling.flood_map import make_flood_map_overlay_v2

class RasterFromTilesLayer(Layer):
    def __init__(self, map, id, map_id, **kwargs):
        super().__init__(map, id, map_id, **kwargs)
        self.new    = True
        self.file_name = map_id + ".png"

    def set_data(self, data):
        self.data = data
        self.update()

    def update(self):
        if not self.get_visibility():
            return
        if self.data is None and self.option == "flood_map":
            return
        overlay_file = os.path.join(self.map.server_path, 'overlays', self.file_name)
        overlay_url  = "./overlays/" + self.file_name
        coords = self.map.map_extent
        xl = [coords[0][0], coords[1][0]]
        yl = [coords[0][1], coords[1][1]]
        wdt = self.map.view.geometry().width()
        hgt = self.map.view.geometry().height()

        # Adjust quality
        if self.quality == "high":
            # No need to do anything (using full screen resolution)
            pass
        elif self.quality == "medium":
            wdt = int(wdt / 2)
            hgt = int(hgt / 2)
        elif self.quality == "low":
            wdt = int(wdt / 4)
            hgt = int(hgt / 4)

        if self.option == "topography":

            xb, yb, caxis = make_topo_overlay_v2(self.topobathy_path,
                                              lon_range=xl,
                                              lat_range=yl,
                                              hillshading=self.hillshading,
                                              hillshading_exaggeration=self.hillshading_exaggeration,
                                              hillshading_azimuth=self.hillshading_azimuth,
                                              hillshading_altitude=self.hillshading_altitude,
                                              npixels=[wdt, hgt],
                                              color_map=self.color_map,
                                              color_range=[self.color_scale_cmin, self.color_scale_cmax],
                                              color_scale_auto=self.color_scale_auto,
                                              color_scale_symmetric=self.color_scale_symmetric,
                                              color_scale_symmetric_side=self.color_scale_symmetric_side,                                       
                                              quiet=False,
                                              file_name=overlay_file)

        elif self.option == "flood_map":
            xb, yb, caxis = make_flood_map_overlay_v2(self.data,
                                        self.index_path,
                                        self.topobathy_path,
                                        npixels=[wdt, hgt],
                                        lon_range=xl,
                                        lat_range=yl,  
                                        option="deterministic",
                                        color_values=self.color_values,
                                        caxis=None,
                                        zbmax=self.zbmax,
                                        depth=None,
                                        quiet=False,
                                        file_name=overlay_file)

        if xb is None:
            # Something went wrong when making the overlay
            return

        # Bounds
        bounds = [[xb[0], xb[1]], [yb[0], yb[1]]]

        # Legend
        if self.color_values:
            clrbar = ColorBar(color_values=self.color_values, legend_title=self.legend_title)
            clrbar.make(0.0, 0.0, decimals=self.decimals)
            clrbar_dict = clrbar.to_dict()

        else:
            # Make colorbar image and send url to js
            # Delete old legend files
            for file_name in glob.glob(os.path.join(self.map.server_path, "overlays", self.map_id + ".legend.*.png")):
                try:
                    os.remove(file_name)
                except:
                    pass

            # add random integer string to legend file to force reload                
            # create string with random integer between 1 and 1,000,000
            rstring = str(np.random.randint(1, 1000000))
            legend_file = self.map_id + ".legend." + rstring + ".png"
            # Multiply cmin and cmax by scale factor
            cmin = caxis[0]
            cmax = caxis[1]
            cmin = cmin * self.scale_factor
            cmax = cmax * self.scale_factor
            cm2png(self.color_map,
                file_name = os.path.join(self.map.server_path, "overlays", legend_file),
                orientation="vertical",
                vmin=cmin,
                vmax=cmax,
                legend_title=self.legend_title,
                legend_label=self.legend_label,
                units=self.legend_units,
                decimals=self.legend_decimals)

            clrbar_dict = "./overlays/" + legend_file

        if self.new:
            self.map.runjs("/js/image_layer.js", "addLayer", arglist=[self.map_id, self.side])
            # Set legend position (should make this generic for all layers)
            self.map.runjs("/js/image_layer.js", "setLegendPosition", arglist=[self.map_id, self.legend_position, self.side])

        self.map.runjs("/js/image_layer.js", "updateLayer", arglist=[overlay_url, self.map_id, bounds, clrbar_dict, self.side])

        self.map.runjs("/js/image_layer.js", "setOpacity", arglist=[self.map_id, self.opacity, self.side])
        self.new = False

    def redraw(self):
        self.new = True
        self.update()
        if not self.get_visibility():
            self.set_visibility(False)



from PyQt5.QtWidgets import QWidget, QDialog
from PyQt5.QtWidgets import QApplication
from PyQt5 import QtGui

class PopupWindow(QDialog):
    def __init__(self, window, modal=True):
        super().__init__()

        self.window = window

        # Determine resize factor for high-res screens
        screen = QApplication.primaryScreen()
        if screen.size().width()>2500:
            self.window.gui.resize_factor = 1.4
        elif screen.size().width()>1200:
            self.window.gui.resize_factor = 1.4
        else:
            self.window.gui.resize_factor = 1.0

        # Window size
        self.setWindowTitle(self.window.title)
        self.window_width = int(self.window.width*self.window.gui.resize_factor)
        self.window_height = int(self.window.height*self.window.gui.resize_factor)
        self.setMinimumSize(self.window_width, self.window_height)
        if self.window.icon:
          self.setWindowIcon(QtGui.QIcon(self.window.icon))
        screen = QApplication.primaryScreen()
        self.setGeometry(int(0.5*(screen.size().width() - self.window_width)),
                         int(0.5*(screen.size().height() - self.window_height)),
                         self.window_width,
                         self.window_height)

        self.setModal(modal)

    def resizeEvent(self, event):
        self.window.resize_elements(self.window.elements)



from .layer import Layer
from geopandas import GeoDataFrame

class PolygonSelectorLayer(Layer):
    def __init__(self, map, id, map_id, **kwargs):
        super().__init__(map, id, map_id, **kwargs)
        pass

    def set_data(self,
                 data,
                 index=None
                 ):
       
        # Make sure this is not an empty GeoDataFrame
        if isinstance(data, GeoDataFrame):
            # Data is GeoDataFrame
            if len(data) == 0:
                data = GeoDataFrame()

        # Set "index" for each feature
        indices = []
        indices.extend(range(len(data)))
        data["index"] = indices
        
        self.data = data 
        
        if isinstance(index, int):
            # Turn index to list to make sure that the feature is selected
            self.index = [index]
        elif index is None:
            # Make sure that no feature is selected
            self.index = []
        else:
            # Index is a list
            self.index = index

        self.map.runjs("/js/polygon_selector_layer.js", "addLayer", arglist=[self.map_id,
                                                                             self.data,
                                                                             self.index,
                                                                             self.hover_property,
                                                                             self.paint_properties,
                                                                             self.selection_type])

        self.select_by_index(self.index)

    def set_selected_index(self, index):
        self.index = index
        self.map.runjs("/js/polygon_selector_layer.js", "setSelectedIndex", arglist=[self.map_id, index])

    # Probably should not have a separate function for this
    # Can also done with layer.hover_property = "name" and then layer.redraw() 
    def set_hover_property(self, hover_property):
        data = self.data
        index = self.index
        self.hover_property = hover_property
        self.set_data(data, index)

    def select_by_index(self, index):
        # index must be a list!
        if isinstance(index, int):
            index = [index]
        self.map.runjs("/js/polygon_selector_layer.js", "selectByIndex", arglist=[self.map_id, index])

    def select_by_id(self, id):
        self.map.runjs("/js/polygon_selector_layer.js", "selectById", arglist=[self.map_id, id])

    def activate(self):
        self.active = True
        if self.data is None:
            return
        self.map.runjs("/js/polygon_selector_layer.js", "activate", arglist=[self.map_id,
                                                                                         self.line_color,
                                                                                         self.fill_color,
                                                                                         self.line_color_selected,
                                                                                         self.fill_color_selected])
  
    def deactivate(self):
        self.active = False
        if self.data is None:
            return
        self.map.runjs("/js/geojson_layer_circle_selector.js", "deactivate", arglist=[self.map_id,
                                                                                          self.line_color_inactive,
                                                                                          self.line_width_inactive,
                                                                                          self.line_style_inactive,
                                                                                          self.line_opacity_inactive,
                                                                                          self.fill_color_inactive,
                                                                                          self.fill_opacity_inactive,
                                                                                          self.line_color_selected_inactive,
                                                                                          self.fill_color_selected_inactive])

    def redraw(self):
        # Called when the map style is changed
        if isinstance(self.data, GeoDataFrame):
            self.set_data(self.data, self.index)
        if not self.get_visibility():
            self.set_visibility(False)

    # def activate(self):
    #     self.active = True

    # def deactivate(self):
    #     self.active = False

    # def clear(self):
    #     self.active = False
    #     self.remove()

    # def remove(self):
    #     self.map.runjs("/js/main.js", "removeLayer", arglist=[self.map_id + ".fill"])
    #     self.map.runjs("/js/main.js", "removeLayer", arglist=[self.map_id + ".line"])
    #     self.map.runjs("/js/main.js", "removeLayer", arglist=[self.map_id])

    # def update(self):
    #     pass

    # def set_visibility(self, true_or_false):
    #     if true_or_false and self.visible:
    #         self.map.runjs("/js/main.js", "showLayer", arglist=[self.map_id + ".fill"])
    #         self.map.runjs("/js/main.js", "showLayer", arglist=[self.map_id + ".line"])
    #     else:
    #         self.map.runjs("/js/main.js", "hideLayer", arglist=[self.map_id + ".fill"])
    #         self.map.runjs("/js/main.js", "hideLayer", arglist=[self.map_id + ".line"])



from PyQt5.QtWidgets import QDateTimeEdit
from PyQt5.QtWidgets import QLabel
from PyQt5 import QtCore
import traceback
import datetime

class DateEdit(QDateTimeEdit):

    def __init__(self, element):
        super().__init__(element.parent.widget)

        self.element = element

        self.setCalendarPopup(True)
        self.setDisplayFormat("yyyy-MM-dd hh:mm:ss")

        if element.text:
            if type(element.text) is str:
                txt = element.text
            else:
                txt = self.element.getvar(element.text.variable_group, element.text.variable)    
            label = QLabel(txt, element.parent.widget)
            label.setStyleSheet("background: transparent; border: none")
            if not element.enable:
                label.setEnabled(False)
            self.text_widget = label
            label.setVisible(True)

        if self.element.tooltip:
            if type(self.element.tooltip) is str:
                txt = self.element.tooltip
            else:
                txt = self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable)    
            self.setToolTip(txt)

        self.editingFinished.connect(self.callback)

        self.set_geometry()

    def set(self):
        group = self.element.variable_group
        name  = self.element.variable
        val   = self.element.getvar(group, name)
        if type(val) is str:
            # Value is a string. Need to convert to datetime using Python datetime.
            val = datetime.datetime.strptime(val, "%Y%m%d %H%M%S")
        dtstr = val.strftime("%Y-%m-%d %H:%M:%S")
        qtDate = QtCore.QDateTime.fromString(dtstr, 'yyyy-MM-dd hh:mm:ss')
        self.setDateTime(qtDate)

        if type(self.element.text) is not str:
            txt = self.element.getvar(self.element.text.variable_group, self.element.text.variable)
            self.text_widget.setText(txt)

        if type(self.element.tooltip) is not str:
            txt = self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable)    
            self.setToolTip(txt)


    def callback(self):
        group = self.element.variable_group
        name  = self.element.variable
        newval = self.dateTime().toPyDateTime()
        if type(self.element.getvar(group, name)) is str:
            # Expected value is a string. Need to convert to newval to string.
            newval = newval.strftime("%Y%m%d %H%M%S")
        # Do some range checking here ...
        # Update value in variable dict
        self.element.setvar(group, name, newval)
        self.okay = True
        try:
            if self.okay and self.isEnabled() and self.element.callback:
                val   = newval
                self.element.callback(val, self)
                # Update GUI
            self.element.window.update()
        except:
            traceback.print_exc()

    def set_geometry(self):
        resize_factor = self.element.gui.resize_factor
        x0, y0, wdt, hgt = self.element.get_position()
        self.setGeometry(x0, y0, wdt, hgt)
        if self.element.text:
            label = self.text_widget
            fm = label.fontMetrics()
            wlab = int(fm.size(0, self.element.text).width())
            if self.element.text_position == "above-center" or self.element.text_position == "above":
                label.setAlignment(QtCore.Qt.AlignCenter)
                label.setGeometry(x0, int(y0 - 20 * resize_factor), wdt, int(20 * resize_factor))
            elif self.element.text_position == "above-left":
                label.setAlignment(QtCore.Qt.AlignLeft)
                label.setGeometry(x0, int(y0 - 20 * resize_factor), wlab, int(20 * resize_factor))
            else:
                # Assuming left
                label.setAlignment(QtCore.Qt.AlignRight)
                label.setGeometry(int(x0 - wlab - 3 * resize_factor),
                                  int(y0 + 5 * self.element.gui.resize_factor),
                                  wlab,
                                  int(20 * resize_factor))



from PySide6.QtWidgets import QCheckBox
from PySide6 import QtCore
import traceback

class CheckBox(QCheckBox):

    def __init__(self, element):
        super().__init__("", element.parent.widget)

        self.element = element

        self.setVisible(True)

        if element.text:
            if isinstance(element.text, str):
                txt = element.text
            else:
                txt = self.element.getvar(element.text.variable_group, element.text.variable)    
            self.setText(txt)    

        if self.element.tooltip:
            if isinstance(self.element.tooltip, str):
                txt = self.element.tooltip
            else:
                txt = self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable)    
            self.setToolTip(txt)

        # self.stateChanged.connect(self.callback)
        self.clicked.connect(self.callback)

        self.set_geometry()

    def set(self):
        group  = self.element.variable_group
        name   = self.element.variable
        val    = self.element.getvar(group, name)
        if val is True:
            self.setChecked(True)
        else:
            self.setChecked(False)
            
        if not isinstance(self.element.text, str):
            self.setText(self.element.getvar(self.element.text.variable_group, self.element.text.variable))   
        if not isinstance(self.element.tooltip, str):
            self.setToolTip(self.element.getvar(self.element.tooltip.variable_group, self.element.tooltip.variable))

    def callback(self, state):
        group = self.element.variable_group
        name = self.element.variable
        val = state
        self.element.setvar(group, name, val)
        try:
            if self.isEnabled() and self.element.callback:
                self.element.callback(val, self)
            # Update GUI
            self.element.window.update()
        except Exception as e:
            print("Error in CheckBox callback")
            print(e)

    def set_geometry(self):
        if self.element.id=="sbg":
            pass
        x0, y0, wdt, hgt = self.element.get_position()
        # Get the width of the text in pixels and add 25 pixels
        txt = self.text()
        wdt = self.fontMetrics().boundingRect(txt).width() + 25
        self.setGeometry(x0, y0, wdt, hgt)




import shapely
from .layer import Layer
from geopandas import GeoDataFrame
import datetime
import copy

class CycloneTrackLayer(Layer):
    def __init__(self, map, id, map_id, **kwargs):
        super().__init__(map, id, map_id, **kwargs)
        pass

    def set_data(self,
                 data):

        # Make sure this is not an empty GeoDataFrame
        if isinstance(data, GeoDataFrame):
            # Data is GeoDataFrame
            if len(data) == 0:
                return
        else:
            print("Data is not a GeoDataFrame")
            return    
        
        # We need to make a copy, because we will add columns for icon_url and hover_html
        self.data = copy.deepcopy(data)

        # Need to add the track to the GeoDataFrame
        # Loop through geometries in GeoDataFrame and make LineString of the points
        # Add this LineString to the GeoDataFrame
        points = []
        # Add columns for icon_url and html
        self.data["icon_url"] = None
        self.data["hover_html"] = None
        for i, row in data.iterrows():
            # Get the point (used to make LineString for new gdf)
            points.append(row["geometry"])
            # Add icon and html to popup
            timestr = row['datetime']
            t = datetime.datetime.strptime(timestr, "%Y%m%d %H%M%S")
            timestr = t.strftime("%Y-%m-%d %H:%M:%S") + " UTC"
            lon = row['geometry'].x
            lat = row['geometry'].y
            vmax = row['vmax']
            pc = row['pc']
            if vmax<64.0:
                cat = "TS"
                icon = "/icons/tropical_storm_icon_24x48.png"
            elif vmax<83.0:
                cat = "1"
                icon = "/icons/category_1_hurricane_icon_24x48.png"
            elif vmax<96.0:    
                cat = "2"
                icon = "/icons/category_2_hurricane_icon_24x48.png"
            elif vmax<113.0:    
                cat = "3"
                icon = "/icons/category_3_hurricane_icon_24x48.png"
            elif vmax<137.0:    
                cat = "4"
                icon = "/icons/category_4_hurricane_icon_24x48.png"
            else:    
                cat = "5"
                icon = "/icons/category_5_hurricane_icon_24x48.png"
            # Make html (&#9; is tab)
            if lat>0.0:
                latstr = f"{abs(lat):.2f}" + " N"
            else:
                latstr = f"{abs(lat):.2f}" + " S"
            if lon>0.0 and lon<180.0:
                lonstr = f"{abs(lon):.2f}" + " E"
            elif lon>180.0:
                lonstr = f"{abs(lon-360.0):.2f}" + " W"    
            else:
                lonstr = f"{abs(lon):.2f}" + " W"    
            vmaxstr = f"{vmax:.1f}"
            if pc>0.0:
                pcstr = f"{pc:.0f}" + " mbar"   
            else:
                pcstr = "N/A"    

            html = 'Time: &#9;' + timestr + '<br />' + \
                   'Category: &#9;' + cat + ' &#9;' + '<br />' + \
                   'Latitude: &#9;' + latstr + ' &#9;' + '<br />' + \
                   'Longitude: &#9;' + lonstr + ' &#9;' + '<br />' + \
                   'Vmax: &#9;' + vmaxstr + ' knots &#9;' + '<br />' + \
                   'Pressure: &#9;' + pcstr + ' &#9;' + '<br />'

            # Add icon and html to row
            self.data.at[i, "icon_size"] = self.icon_size
            self.data.at[i, "icon_url"] = icon
            self.data.at[i, "hover_html"] = html

        # Create shapely LineString with geometries in points
        line = shapely.geometry.LineString(points)
        # Now make new gdf with LineString
        track_gdf = GeoDataFrame([{"geometry": line}]).set_crs(4326)

        # This is a composite layer with a line and icons

        # First add the line layer
        id = self.map_id + ".track_line"
        self.map.runjs("/js/line_layer.js", "addLayer", arglist=[id,
                                                                             track_gdf,
                                                                             self.line_color,
                                                                             self.line_width,
                                                                             self.line_opacity,
                                                                             self.fill_color,
                                                                             self.fill_opacity,
                                                                             0])

        if self.show_icons:
            # First add the line layer
            id = self.map_id + ".track_points"
            self.map.runjs("/js/marker_layer.js", "addLayer", arglist=[id, self.data.to_crs(4326)])

    def redraw(self):
        if isinstance(self.data, GeoDataFrame):
            self.set_data(self.data)
        if not self.get_visibility:
            self.set_visibility(False)

    # TODO: Implement activate and deactivate
    # def activate(self):
    #     self.map.runjs("/js/geojson_layer_line.js", "setPaintProperties", arglist=[self.map_id,
    #                                                                                    self.line_color,
    #                                                                                    self.line_width,
    #                                                                                    self.line_opacity,
    #                                                                                    self.fill_color,
    #                                                                                    self.fill_opacity,
    #                                                                                    self.circle_radius])
  
    # def deactivate(self):
    #     self.map.runjs("/js/geojson_layer_line.js", "setPaintProperties", arglist=[self.map_id,
    #                                                                                    self.line_color_inactive,
    #                                                                                    self.line_width_inactive,
    #                                                                                    self.line_opacity_inactive,
    #                                                                                    self.fill_color_inactive,
    #                                                                                    self.fill_opacity_inactive,
    #                                                                                    self.circle_radius_inactive])

    def set_visibility(self, true_or_false):
        # This is a composite layer. set_visibility overrides method in Layer class
        if true_or_false:
            self.map.runjs(self.main_js, "showLayer", arglist=[self.map_id + ".track_line", self.side])
            self.map.runjs(self.main_js, "showLayer", arglist=[self.map_id + ".track_points", self.side])
        else:
            self.map.runjs(self.main_js, "hideLayer", arglist=[self.map_id + ".track_line", self.side])
            self.map.runjs(self.main_js, "hideLayer", arglist=[self.map_id + ".track_points", self.side])

    def delete_from_map(self):
        self.map.runjs(self.main_js, "removeLayer", arglist=[self.map_id + ".track_line", self.side])
        self.map.runjs(self.main_js, "removeLayer", arglist=[self.map_id + ".track_points", self.side])

    def clear(self):
        self.data = None
        self.delete_from_map()



from PySide6.QtWidgets import QLabel, QPushButton, QGroupBox
from PySide6 import QtCore, QtGui
import traceback

# class Frame(QFrame):
class Frame(QGroupBox):
    def __init__(self, element):
        super().__init__(element.parent.widget)

        self.element = element

        collapsable = False

        if element.collapse:
            # Parent
            collapsable = True
            # Add pushbutton to collapse
            self.pushbutton = QPushButton("", self)
            self.pushbutton.clicked.connect(self.collapse_callback)

        if hasattr(element.parent, "style"):
            if element.parent.style == "panel" and element.parent.collapse:
                collapsable = True

        if collapsable:
            # self.setLineWidth(0)
            pass
        else:
            # Regular
            # self.setFrameShape(QFrame.StyledPanel)
            # self.setTitle(element.text)
            # self.setStyleSheet("background-color: transparent; border: 2px solid white;")
            # self.setFrameShape(QFrame.Box)
            # self.setLineWidth(2)
            # self.setLineWidth(1)
            pass

        if element.text:
            self.text_widget = QLabel(element.text, element.parent.widget)
            # font = QtGui.QFont()
            # font.setWeight(QtGui.QFont.Weight.Bold)
            # self.text_widget.setFont(font)

        self.set_geometry()

    def set(self):
        pass

    def showEvent(self, event):
        # What is this for? It was added by Daley at some point.
        # Setting the geometry at this point makes frames disappear because the size of the parent is 0,0,-1,-1 for some reason.
        # Comment out for now
        # self.set_geometry()
        pass

    def set_geometry(self):

        rf = self.element.gui.resize_factor
        button_size = 12
        x0, y0, wdt, hgt = self.element.get_position()
        # if self.element.text:
        #     hgt = hgt + 20
            # self.setStyleSheet("margin-top: 10px;")
        if self.element.collapse:
            wdt = wdt + button_size
        self.setGeometry(x0, y0, wdt, hgt)

        # Text widget
        if self.element.text:
            fm = self.text_widget.fontMetrics()
            hlab = fm.size(0, self.element.text).height()
            wlab = fm.size(0, self.element.text).width()
            self.text_widget.setGeometry(int(x0 + 5*rf), int(y0 - 0.6*hlab), wlab, hlab)
            self.text_widget.setAlignment(QtCore.Qt.AlignTop)

        # Push button
        button_size = 16
        if self.element.collapse:
            wdtp = button_size
            hgtp = button_size
            x0p = wdt - button_size
            y0p = int(0.5 * hgt - 0.5 * hgtp)
            self.pushbutton.setGeometry(x0p, y0p, wdtp, hgtp)

        # Check if these are collapable panels
        if hasattr(self.element.parent, "style"):
            if self.element.parent.style == "panel" and self.element.parent.collapse:
                pwdt = self.element.parent.widget.geometry().width() - button_size
                phgt = self.element.parent.widget.geometry().height()
                if self.element.parent.collapsed:
                    arrow_file = "icons8-triangle-arrow-16_white_left.png"
                else:
                    arrow_file = "icons8-triangle-arrow-16_white_right.png"                    
                self.element.parent.widget.pushbutton.setStyleSheet(
                    "background-image : url(:/img/" + arrow_file + "); border: none")

                if self.element == self.element.parent.elements[0]:
                    # First panel
                    x0 = 0
                    y0 = 0
                    if self.element.parent.collapsed:
                        wdt = int(self.element.parent.fraction_collapsed * pwdt)
                    else:
                        wdt = int(self.element.parent.fraction_expanded * pwdt)
                    hgt = phgt
                else:
                    # Second panel
                    if self.element.parent.collapsed:
                        x0 = int(self.element.parent.fraction_collapsed * pwdt)
                    else:
                        x0 = int(self.element.parent.fraction_expanded * pwdt)
                    y0 = 0
                    wdt = pwdt - x0
                    hgt = phgt
                self.setGeometry(x0, y0, wdt, hgt)

    def collapse_callback(self):
        try:
            if self.element.collapsed:
                self.element.collapsed = False
            else:
                self.element.collapsed = True
            self.element.gui.window.resize_elements(self.element.parent.elements)
            if self.element.callback:
                self.element.callback(self)
                # Update GUI
                self.element.window.update()
        except:
            traceback.print_exc()



from .layer import Layer
from geopandas import GeoDataFrame

class CircleSelectorLayer(Layer):
    def __init__(self, map, id, map_id, **kwargs):
        super().__init__(map, id, map_id, **kwargs)
        pass

    def set_data(self,
                 data,
                 index,
                 crs=None):

        self.data = data
        self.index = index

        # # Remove existing layer
        # self.map.runjs("/js/main.js", "removeLayer", arglist=[self.map_id])

        # Make sure this is not an empty GeoDataFrame
        if isinstance(data, GeoDataFrame):
            # Data is GeoDataFrame
            if len(data) == 0:
                self.clear()
                return

        indices = []
        indices.extend(range(len(data)))
        data["index"] = indices

        # Add new layer
        self.map.runjs("/js/circle_selector_layer.js", "addLayer", arglist=[self.map_id,
                                                                                data.to_crs(4326),
                                                                                index,
                                                                                self.hover_property,
                                                                                self.line_color,
                                                                                self.line_width,
                                                                                self.line_style,
                                                                                self.line_opacity,
                                                                                self.fill_color,
                                                                                self.fill_opacity,
                                                                                self.circle_radius,
                                                                                self.line_color_selected,
                                                                                self.fill_color_selected,
                                                                                self.circle_radius_selected,
                                                                                self.selection_type])

    def select_by_index(self, index):
        self.map.runjs("/js/circle_selector_layer.js", "selectByIndex", arglist=[self.map_id, index])

    def select_by_id(self, id):
        self.map.runjs("/js/circle_selector_layer.js", "selectById", arglist=[self.map_id, id])

    def activate(self):
        if self.data is None:
            return
        elif len(self.data) == 0:
            return
        self.show()
        self.map.runjs("/js/circle_selector_layer.js", "activate", arglist=[self.map_id,
                                                                                        self.line_color,
                                                                                        self.line_width,
                                                                                        self.line_style,
                                                                                        self.line_opacity,
                                                                                        self.fill_color,
                                                                                        self.fill_opacity,
                                                                                        self.circle_radius,
                                                                                        self.line_color_selected,
                                                                                        self.fill_color_selected,
                                                                                        self.circle_radius_selected])
  
    def deactivate(self):
        if self.data is None:
            return
        elif len(self.data) == 0:
            return
        self.map.runjs("/js/circle_selector_layer.js", "deactivate", arglist=[self.map_id,
                                                                                        self.line_color_inactive,
                                                                                        self.line_width_inactive,
                                                                                        self.line_style_inactive,
                                                                                        self.line_opacity_inactive,
                                                                                        self.fill_color_inactive,
                                                                                        self.fill_opacity_inactive,
                                                                                        self.circle_radius_inactive,
                                                                                        self.line_color_inactive,
                                                                                        self.fill_color_inactive,
                                                                                        self.circle_radius_inactive])

    def redraw(self):
        if isinstance(self.data, GeoDataFrame):
            self.set_data(self.data, self.index)
        if not self.get_visibility():
            self.set_visibility(False)



import math
import pandas as pd
import geopandas as gpd
import matplotlib.colors as mcolors
import shapely
from shapely.geometry import Polygon
from shapely.ops import transform
import pyproj

from .layer import Layer


class DrawLayer(Layer):
    def __init__(
        self,
        map,
        layer_id,
        map_id,
        columns={},
        shape="polygon",
        create=None,
        modify=None,
        select=None,
        deselect=None,
        add=None,
        rotate=True,
        polygon_line_color="dodgerblue",
        polygon_line_width=2,
        polygon_line_style="-",
        polygon_line_opacity=1.0,
        polygon_fill_color="dodgerblue",
        polygon_fill_opacity=0.5,
        polyline_line_color="limegreen",
        polyline_line_width=2,
        polyline_line_style="-",
        polyline_line_opacity=1.0,
        circle_line_color="black",
        circle_line_opacity=1.0,
        circle_fill_color="orangered",
        circle_fill_opacity=0.5,
        circle_radius=4,
    ):
        super().__init__(map, layer_id, map_id)

        self.active = False
        self.type = "draw"
        self.shape = shape
        self.mode = (
            "active"  # Draw layers can have three modes: active, inactive, invisible
        )
        self.gdf = gpd.GeoDataFrame()
        self.columns = columns
        self.create = create
        self.modify = modify
        self.select = select
        self.deselect = deselect
        self.add = add

        # Get Hex values for colors
        self.paint_props = {}
        self.paint_props["polygon_line_color"] = mcolors.to_hex(polygon_line_color)
        self.paint_props["polygon_line_width"] = polygon_line_width
        self.paint_props["polygon_line_style"] = polygon_line_style
        self.paint_props["polygon_line_opacity"] = polygon_line_opacity
        self.paint_props["polygon_fill_color"] = mcolors.to_hex(polygon_fill_color)
        self.paint_props["polygon_fill_opacity"] = polygon_fill_opacity
        self.paint_props["polyline_line_color"] = mcolors.to_hex(polyline_line_color)
        self.paint_props["polyline_line_width"] = polyline_line_width
        self.paint_props["polyline_line_style"] = polyline_line_style
        self.paint_props["polyline_line_opacity"] = polyline_line_opacity
        self.paint_props["circle_line_color"] = mcolors.to_hex(circle_line_color)
        self.paint_props["circle_line_opacity"] = circle_line_opacity
        self.paint_props["circle_fill_color"] = mcolors.to_hex(circle_fill_color)
        self.paint_props["circle_fill_opacity"] = circle_fill_opacity
        self.paint_props["circle_radius"] = circle_radius
        self.paint_props["rotate"] = rotate

        # Add this layer
        self.map.runjs(
            "/js/draw_layer.js",
            "addLayer",
            arglist=[self.map_id, "active", self.paint_props, self.shape],
        )

    def set_mode(self, mode):
        """Set mode of layer. Can be active, inactive or invisible."""
        self.mode = mode
        if mode == "active":
            self.active = True
            self.visible = True
        elif mode == "inactive":
            self.active = False
            self.visible = True
        else:
            self.active = False
            self.visible = False
        self.map.runjs("/js/draw_layer.js", "setLayerMode", arglist=[self.map_id, mode])

    def activate(self):
        """Activate the draw layer so it can be edited by the user."""
        self.visible = True
        self.active = True
        self.map.runjs(
            "/js/draw_layer.js", "setLayerMode", arglist=[self.map_id, "active"]
        )

    def deactivate(self):
        """Deactivate the draw layer so it can not be edited by the user."""
        self.active = False
        self.map.runjs(
            "/js/draw_layer.js", "setLayerMode", arglist=[self.map_id, "inactive"]
        )

    # def show(self):
    #     self.visible = True
    #     if self.active:
    #         self.map.runjs("/js/draw_layer.js", "setLayerMode", arglist=[self.map_id, "active"])
    #     else:
    #         self.map.runjs("/js/draw_layer.js", "setLayerMode", arglist=[self.map_id, "inactive"])

    # def hide(self):
    #     self.visible = False
    #     self.map.runjs("/js/draw_layer.js", "setLayerMode", arglist=[self.map_id, "invisible"])

    # def set_mode(self, mode):
    #     self.mode = mode
    #     self.map.runjs("/js/draw_layer.js", "setLayerMode", arglist=[self.map_id, mode])

    def set_visibility(self, true_or_false):
        if true_or_false:
            if self.active:
                self.map.runjs(
                    "/js/draw_layer.js", "setLayerMode", arglist=[self.map_id, "active"]
                )
            else:
                self.map.runjs(
                    "/js/draw_layer.js",
                    "setLayerMode",
                    arglist=[self.map_id, "inactive"],
                )
        else:
            # Make layer invisible
            self.map.runjs(
                "/js/draw_layer.js", "setLayerMode", arglist=[self.map_id, "invisible"]
            )

    def set_data(self, gdf):
        """Clear the draw layer and add data. Data must be a GeoDataFrame."""
        self.clear()
        self.add_feature(gdf)

    def add_feature(self, gdf):
        """Add data to draw layer. Data must be a GeoDataFrame."""

        if len(gdf) == 0:
            return

        # Append to self.gdf (which doesn't have to have CRS=4326)
        if not self.gdf.empty:
            self.gdf = pd.concat(
                [self.gdf, gdf.to_crs(self.gdf.crs)], ignore_index=True
            )
        else:
            self.gdf = gdf

        for index, row in gdf.to_crs(4326).iterrows():
            # Add feature to draw layer, which has CRS=4326
            gdf2plot = gpd.GeoDataFrame(geometry=[row["geometry"]])
            # if "name" in row:
            #     gdf["name"] = row["name"]
            self.map.runjs(
                "/js/draw_layer.js", "addFeature", arglist=[gdf2plot, self.map_id]
            )

    def add_rectangle(self, x0, y0, lenx, leny, rotation):
        """Add rectangle to draw layer."""
        x = [x0]
        y = [y0]
        x.append(x[0] + lenx * math.cos(math.pi * rotation / 180))
        y.append(y[0] + lenx * math.sin(math.pi * rotation / 180))
        x.append(x[1] + leny * math.cos(math.pi * (rotation + 90.0) / 180))
        y.append(y[1] + leny * math.sin(math.pi * (rotation + 90.0) / 180))
        x.append(x[2] + lenx * math.cos(math.pi * (rotation + 180.0) / 180))
        y.append(y[2] + lenx * math.sin(math.pi * (rotation + 180.0) / 180))
        x.append(x0)
        y.append(y0)
        polygon_crs = Polygon(zip(x, y))
        # Transform to WGS 84
        project = pyproj.Transformer.from_crs(
            self.map.crs, pyproj.CRS(4326), always_xy=True
        ).transform
        polygon_wgs84 = transform(project, polygon_crs)
        gdf = gpd.GeoDataFrame(geometry=[polygon_wgs84], crs=4326)
        gdf["x0"] = x0
        gdf["y0"] = y0
        gdf["dx"] = lenx
        gdf["dy"] = leny
        gdf["rotation"] = rotation
        self.add_feature(gdf)

    def draw(self):
        """Activate drawing mode."""
        if self.shape == "polygon":
            self.map.runjs("/js/draw_layer.js", "drawPolygon", arglist=[self.map_id])
        elif self.shape == "polyline":
            self.map.runjs("/js/draw_layer.js", "drawPolyline", arglist=[self.map_id])
        elif self.shape == "rectangle":
            self.map.runjs("/js/draw_layer.js", "drawRectangle", arglist=[self.map_id])

    def set_gdf(self, feature_collection, index):

        # Called after a feature has been drawn or added. A new GeoDataFrame is created for this layer.
        gdf = gpd.GeoDataFrame.from_features(feature_collection, crs=4326).to_crs(
            self.map.crs
        )

        feature_id = feature_collection["features"][index]["id"]

        if index > len(self.gdf) - 1:
            # Add new row to self.gdf
            d = {"geometry": [gdf.iloc[index]["geometry"]], "id": [feature_id]}
            if self.columns:
                for col in self.columns:
                    d[col] = [self.columns[col]]
            new_row = gpd.GeoDataFrame(d).set_crs(self.map.crs)
            # Add new row to self.gdf (this means that the dataframe will get a new memory address)
            self.gdf = pd.concat([self.gdf, new_row], ignore_index=True)
            pass
        else:
            # Only change the geometry of the selected feature and copy the id
            # self.gdf.at[index, "id"] = gdf.at[index, "id"]
            self.gdf.at[index, "id"] = feature_id
            self.gdf.at[index, "geometry"] = gdf.at[index, "geometry"]

    def get_feature_index(self, feature_id):
        """Get a features index by ID in the GeoDataFrame."""
        feature_index = None
        if len(self.gdf) > 0:
            indx = self.gdf.index[self.gdf["id"] == feature_id].tolist()
            if indx:
                feature_index = indx[0]
        if feature_index is None:
            print("Could not find feature ...")
        return feature_index

    def get_feature_id(self, feature_index):
        """Get a features ID by index in the GeoDataFrame."""
        feature_id = None
        if len(self.gdf) > 0 and len(self.gdf) >= feature_index + 1:
            feature_id = self.gdf.loc[feature_index, "id"]
        return feature_id

    def feature_drawn(self, feature_collection, feature_id):

        # Feature index is always the last in the feature collection
        feature_index = len(feature_collection["features"]) - 1

        # Called after a feature has been drawn by the user
        if self.shape == "rectangle":
            # Need to make sure that the rectangle is drawn in ccw direction
            # with the lower left corner as the first point
            geom = feature_collection["features"][feature_index]["geometry"]
            geom = fix_rectangle_geometry(geom)
            feature_collection["features"][feature_index]["geometry"] = geom

        # Create geodataframe from feature collection
        self.set_gdf(feature_collection, feature_index)

        # Check if this is a rectangle. If so, we need to compute the geometry and redraw it,
        # since the polygon may have been changed.
        if self.shape == "rectangle":
            # Now compute geometry for rectangle that was just drawn
            # Not geom is in the map CRS, not necessarily 4326
            geom = self.gdf.loc[feature_index, "geometry"]
            x0, y0, dx, dy, rotation = get_rectangle_geometry(geom)
            self.gdf.at[feature_index, "x0"] = x0
            self.gdf.at[feature_index, "y0"] = y0
            self.gdf.at[feature_index, "dx"] = dx
            self.gdf.at[feature_index, "dy"] = dy
            self.gdf.at[feature_index, "rotation"] = 0.0 # When drawing a rectangle, the rotation is always 0.0
            # Need to set the corrected rectangle geometry, but it has to be in 4326
            # Transform to WGS 84
            project = pyproj.Transformer.from_crs(
                self.map.crs, pyproj.CRS(4326), always_xy=True
            ).transform
            geom = transform(project, geom) 
            self.set_feature_geometry(feature_id, geom)

        # Check if there is a create method
        if self.create:
            # Call the create method sending the gdf with all features and the index of the new feature
            self.create(self.gdf, feature_index, feature_id)

    def feature_added(self, feature_collection, feature_id):
        # Called after a feature has been added by the application
        index = len(feature_collection["features"]) - 1
        self.set_gdf(feature_collection, index)
        if self.add:
            feature_index = self.get_feature_index(feature_id)
            self.add(self.gdf, feature_index, feature_id)

    def feature_modified(self, feature_collection, feature_id):
        # Called after a feature has been modified by the user
        index = self.get_feature_index(feature_id)
        self.set_gdf(feature_collection, index)

        if self.shape == "rectangle":
            # Need to set rectangle geometry
            geom = self.gdf.loc[index, "geometry"]
            x0, y0, dx, dy, rotation = get_rectangle_geometry(geom)
            self.gdf.at[index, "x0"] = x0
            self.gdf.at[index, "y0"] = y0
            self.gdf.at[index, "dx"] = dx
            self.gdf.at[index, "dy"] = dy
            self.gdf.at[index, "rotation"] = rotation

        if self.modify:
            self.modify(self.gdf, index, feature_id)

    def feature_selected(self, feature_collection, feature_id):
        # Called after a feature has been selected by the user
        if self.select:
            feature_index = self.get_feature_index(feature_id)
            self.select(feature_index)

    def feature_deselected(self):
        # Called after a feature has been de-selected by the user
        if self.deselect:
            self.deselect()

    def activate_feature(self, feature_id_or_index):
        """Activate a feature by ID or index so it can be edited by the user."""
        if isinstance(feature_id_or_index, int):
            # It's an index
            index = feature_id_or_index
            feature_id = self.get_feature_id(index)
        else:
            feature_id = feature_id_or_index
        if self.mode != "active":
            self.mode = "active"
            self.map.runjs(
                "/js/draw_layer.js", "setLayerMode", arglist=[self.map_id, self.mode]
            )
        self.map.runjs("/js/draw_layer.js", "activateFeature", arglist=[feature_id])

    def set_feature_geometry(self, feature_id, geom):
        # This does not work!
        if hasattr(geom, "__geo_interface__"):
            # Shapely geometry, turn into dict
            geom = geom.__geo_interface__
        self.map.runjs(
            "/js/draw_layer.js",
            "setFeatureGeometry",
            arglist=[self.map_id, feature_id, geom],
        )

    def delete_feature(self, feature_id_or_index):
        """Delete a feature by ID or index"""
        if isinstance(feature_id_or_index, int):
            # It's an index
            index = feature_id_or_index
            feature_id = self.get_feature_id(index)
            # Remove from gdf
            self.gdf = self.gdf.drop(index)
            if len(self.gdf) > 0:
                self.gdf = self.gdf.reset_index(drop=True)
            else:
                self.gdf = gpd.GeoDataFrame()
            # Remove from map
            self.map.runjs("/js/draw_layer.js", "deleteFeature", arglist=[feature_id])
        else:
            feature_id = feature_id_or_index
            # Remove from gdf
            for index, row in self.gdf.iterrows():
                if row["id"] == feature_id:
                    self.gdf = self.gdf.drop(index)
                    if len(self.gdf) > 0:
                        self.gdf = self.gdf.reset_index(drop=True)
                    else:
                        self.gdf = gpd.GeoDataFrame()
                    break
            self.map.runjs("/js/draw_layer.js", "deleteFeature", arglist=[feature_id])
        return self.gdf

    def delete_from_map(self):
        """Delete the draw layer from the map."""
        self.active = False
        self.map.runjs("/js/draw_layer.js", "deleteLayer", arglist=[self.map_id])
        self.gdf = gpd.GeoDataFrame()

    def clear(self):
        """Clear the draw layer."""
        for index, row in self.gdf.iterrows():
            self.map.runjs("/js/draw_layer.js", "deleteFeature", arglist=[row["id"]])
        self.gdf = gpd.GeoDataFrame()

    def get_gdf(self, id=None):
        """Return the GeoDataFrame for this layer. If id is specified, return only the row with that id."""
        if id:
            for index, row in self.gdf.iterrows():
                if row["id"] == id:
                    return self.gdf[index]
        else:
            return self.gdf

    def redraw(self):
        # Add this layer
        self.map.runjs(
            "/js/draw_layer.js",
            "addLayer",
            arglist=[self.map_id, "active", self.paint_props, self.shape],
        )
        self.add_feature(self.gdf)
        self.set_mode(self.mode)
        if not self.get_visibility():
            self.set_visibility(False)


def get_rectangle_geometry(geom):
    xx, yy = geom.exterior.coords.xy
    x0 = float(xx[0])
    y0 = float(yy[0])
    dx = math.sqrt(float(xx[1] - xx[0]) ** 2 + float(yy[1] - yy[0]) ** 2)
    dy = math.sqrt(float(xx[2] - xx[1]) ** 2 + float(yy[2] - yy[1]) ** 2)
    rotation = float((math.atan2(yy[1] - yy[0], xx[1] - xx[0])))
    if abs(rotation * 180 / math.pi) < 1.0:
        rotation = 0.0
    return x0, y0, dx, dy, rotation


def fix_rectangle_geometry(geom):
    # This ensures that the rectangle is drawn CCW direction with the lower left corner as the first point (what a mess ...)
    # Check if the geometry is a polygon
    # Check if this ccw
    if not shapely.is_ccw(shapely.LineString(geom["coordinates"][0])):
        # Reverse the order of the coordinates
        geom["coordinates"][0] = geom["coordinates"][0][::-1]
    coords = geom["coordinates"][0]
    # Remove the last point which is the same as the first
    coords = coords[0:-1]
    # Get lower left corner by taking minimum x and y values
    min_x = min([x for x, y in coords])
    min_y = min([y for x, y in coords])
    # Now find the index of the point closest to the lower left corner
    dst = [(x - min_x) ** 2 + (y - min_y) ** 2 for x, y in coords]
    index = dst.index(min(dst))
    new_first_point = coords[
        index
    ]  # Assuming you want the third point as the new first point
    new_coords = [new_first_point] + coords[index + 1 :] + coords[0:index]
    new_coords.append(new_coords[0])
    # Create a new polygon with the updated coordinates
    new_polygon = Polygon(new_coords)
    # Update the geometry
    geom = new_polygon.__geo_interface__
    return geom



from PySide6.QtWidgets import QStatusBar, QLabel

class StatusBar:
    def __init__(self, window):
        self.parent = window.window_widget
        sb = QStatusBar()
        self.parent.setStatusBar(sb)
        self.widget = sb
        field_list = window.statusbar_fields
        self.field  = {}
        for field in field_list:
            id = field["id"]
            w = field["width"]
            text = field["text"]
            label = QLabel(text)
            # label.setStyleSheet("border :2px solid blue;") 
            self.field[id] = label
            sb.addPermanentWidget(label, w)
        # self.set_geometry()    

    def set_geometry(self):
        return

    def show_message(self, message, time):
        self.widget.showMessage(message, time)        

    def set_text(self, id, text):
        self.field[id].setText(text)



from .layer import Layer
from geopandas import GeoDataFrame


class PolygonLayer(Layer):
    def __init__(self, map, id, map_id, **kwargs):
        super().__init__(map, id, map_id, **kwargs)
        pass

    def set_data(
        self,
        data,
        index=None,
    ):
        # Index is not used, but is kept for consistency with polygon_selector_layer

        # Make sure this is not an empty GeoDataFrame
        if isinstance(data, GeoDataFrame):
            # Data is GeoDataFrame
            if len(data) == 0:
                data = GeoDataFrame()

        self.data = data

        # Add new layer
        self.map.runjs(
            "/js/simple_polygon_layer.js",
            "addLayer",
            arglist=[
                self.map_id,
                self.data.to_crs(4326),
                self.line_color,
                self.line_width,
                self.line_opacity,
                self.line_style,
            ],
        )

    def activate(self):
        self.active = True
        if self.data is None:
            return
        self.map.runjs(
            "/js/simple_polygon_layer.js",
            "activate",
            arglist=[self.map_id, self.line_color],
        )

    def deactivate(self):
        self.active = False
        if self.data is None:
            return
        self.map.runjs(
            "/js/geojson_layer_circle.js", # WTF is this?
            "deactivate",
            arglist=[
                self.map_id,
                self.line_color_inactive,
            ],
        )

    def redraw(self):
        # Called when the map style is changed
        if isinstance(self.data, GeoDataFrame):
            self.set_data(self.data)
        if not self.get_visibility():
            self.set_visibility(False)


import os
import yaml
import sys
import copy
import shutil
import importlib

from pathlib import Path
import toml

from guitares.window import Window
from guitares.server import start_server

class GUI:
    def __init__(self, module,
                 framework="pyqt5",
                 splash_file=None,
                 stylesheet=None,
                 config_path=None,
                 config_file=None,
                 icon=None,
                 server_path=None,
                 server_port=3000,
                 server_nodejs=False,
                 js_messages=True,
                 copy_map_server_folder=True,
                 icon_path=None,
                 map_engine="mapbox",
                 mapbox_token_file="mapbox_token.txt"):

        self.module      = module
        self.framework   = framework
        self.splash_file = splash_file
        self.stylesheet  = stylesheet
        self.config_file = config_file
        self.config_path = config_path
        self.map_engine  = map_engine
        self.splash      = None
        self.icon        = icon
        self.config      = {}
        self.variables   = {}
        self.server_thread = None
        self.server_path = server_path
        self.server_port = server_port
        self.server_thread = None
        self.server_nodejs = server_nodejs
        self.js_messages = js_messages
        self.popup_window = {}
        self.popup_data   = {}
        self.resize_factor = 1.0

        # os.environ["QT_LOGGING_RULES"] = "*.debug=false;qt.webchannel.*=false"            

        if self.framework == "pyqt5":
            from PyQt5.QtWidgets import QApplication
            from PyQt5 import QtCore
            import guitares.pyqt5.icons_rc
        elif self.framework == "pyside6":
            from PySide6.QtWidgets import QApplication
            from PySide6 import QtCore
            import guitares.pyside6.icons_rc

        QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
        self.qtapp = QApplication(sys.argv)

        # Show splash screen
        self.show_splash()

        if not self.config_path:
            self.config_path = os.getcwd()

        if self.config_file:
            # Read configuration file  
            self.config = self.read_gui_config(self.config_path, self.config_file)
        else:
            # No configuration file, so just set config to empty dict, which MUST be filled by the developer before calling build
            self.config = {}

        self.image_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "img")
        self.image_path = self.image_path.replace(os.sep, '/')

        if server_path:
            # Need to run http server (e.g. for MapBox or MapLibre)
            print("Starting http server ...")
            # Run http server in separate thread
            # Use daemon=True to make sure the server stops after the application is finished
            if copy_map_server_folder:
                mppth = os.path.join(os.path.dirname(os.path.abspath(__file__)), "map", self.map_engine, "server")
                # Delete current server folder
                if os.path.exists(server_path):
                    shutil.rmtree(server_path)
                # Now copy over folder from mapbox or maplibre
                shutil.copytree(mppth, server_path)

            # Check if mapbox token file exists in config path or in current working directory
            token_file_name = None    
            if os.path.exists(os.path.join(self.config_path, mapbox_token_file)):
                token_file_name = os.path.join(self.config_path, mapbox_token_file)
            if os.path.exists(os.path.join(os.getcwd(), mapbox_token_file)):
                token_file_name = os.path.join(os.getcwd(), mapbox_token_file)    

            # Read mapbox token and store in js file in server path
            if token_file_name:
                fid = open(token_file_name, "r")
                mapbox_token = fid.readlines()
                fid.close()
                fid = open(os.path.join(server_path, "mapbox_token.js"), "w")
                fid.write("mapbox_token = '" + mapbox_token[0].strip() + "';")
                fid.close()

            if icon_path:
                # Copy all files in icon_path to server_path/icons
                if not os.path.exists(os.path.join(server_path, "icons")):
                    os.mkdir(os.path.join(server_path, "icons"))
                for file in os.listdir(icon_path):
                    if file.endswith(".png"):
                        shutil.copy(os.path.join(icon_path, file), os.path.join(server_path, "icons", file))
            
            # Set map styles
            if self.map_engine == "mapbox":
                self.map_style = "mapbox://styles/mapbox/streets-v12"
            elif self.map_engine == "maplibre":
                self.map_style = "osm"
#                self.map_styles = ["osm", "orto", "darkmatter", "positron", "none"]

            start_server(server_path, port=server_port, node=self.server_nodejs)

    def show_splash(self):
        if self.splash_file:
            mod = importlib.import_module(f"guitares.{self.framework}.splash")
            self.splash = mod.Splash(self.splash_file, seconds=20.0).splash

    def close_splash(self):
        if self.splash:
            self.splash.close()

    def build(self):

        app = self.qtapp

        # Set the icon
        if self.framework == "pyqt5":
            from PyQt5 import QtGui
        elif self.framework == "pyside6":
            from PySide6 import QtGui

        app.setWindowIcon(QtGui.QIcon(self.icon))

        if self.stylesheet:
            app.setStyleSheet(open(os.path.join(self.config_path, self.stylesheet), "r").read())

        # if self.config_file:
        #     self.config = self.read_gui_config(self.config_path, self.config_file)

        # Make window object
        self.window = Window(self.config, self)
        self.window.build()

        # Call on_build method after building window
        if hasattr(self.module, "on_build"):
            self.module.on_build()
            
        # # # Close splash screen before GUI is initiated

        app.exec_()

    def setvar(self, group, name, value):
        if group not in self.variables:
            self.variables[group] = {}
        if name not in self.variables[group]:
            self.variables[group][name] = {}
        self.variables[group][name]["value"] = value

    def getvar(self, group, name):
        if name is None:
            # There is no variable for the element
            return None
        if group not in self.variables:
            print("Error! Cannot get variable! GUI variable group '" + group + "' not defined!")
            return None
        elif name not in self.variables[group]:
            print("Error! Cannot get variable! GUI variable '" + name + "' not defined in group '" + group + "'!")
            return None
        return self.variables[group][name]["value"]
    
    def delvar(self, group, name):
        if group not in self.variables:
            print("Error! Cannot delete variable! GUI variable group '" + group + "' not defined!")
            return None
        elif name not in self.variables[group]:
            print("Error! Cannot delete variable! GUI variable '" + name + "' not defined in group '" + group + "'!")
            return None
        del self.variables[group][name]

    def delgroup(self, group):
        if group not in self.variables:
            print("Error! Cannot delete group! GUI variable group '" + group + "' not defined!")
            return
        del self.variables[group]
        
    def popup(self, config, id="popup", data=None):
        # Make pop-up window
        # config needs to be file name of yml file, or configuration dict
        # Data is optional and can have any shape (e.g. dict, str, object, etc.)
        # Data will only be changed if Okay is clicked in the pop-up window
        if isinstance(config, str):
            path = os.path.dirname(config)
            file_name = os.path.basename(config)
            config = self.read_gui_config(path, file_name)
        if data:    
            self.popup_data[id] = copy.copy(data)
        else:
            self.popup_data[id] = None    
        self.popup_window[id] = Window(config, self, type="popup")
        p = self.popup_window[id].build()
        okay = False
        if p.result() == 1:
            okay = True
            data = self.popup_data[id]
        # Remove popup window and data    
        return okay, data

    def read_gui_config(self, path, file_name):
        suffix = Path(path).joinpath(file_name).suffix
        if suffix == '.yml':
            d = yaml2dict(os.path.join(path, file_name))
        elif suffix == '.toml':
            d = toml.load(os.path.join(path, file_name))
        config = {}
        config["window"] = {}
        config["toolbar"] = {}
        config["menu"] = []
        config["element"] = []
        config["statusbar"] = {}
        if "window" in d:
            config["window"] = d["window"]
        if "toolbar" in d:
            config["toolbar"] = d["toolbar"]
        if "menu" in d:
            # Recursively read menu
            config["menu"] = d["menu"]
        if "element" in d:
            # Recursively read elements
            config["element"] = self.read_gui_elements(path, file_name)
        if "statusbar" in d:
            config["statusbar"] = d["statusbar"]
        return config

    def read_gui_elements(self, path, file_name):
        # Return just the elements
        suffix = Path(path).joinpath(file_name).suffix
        if suffix == '.yml':
            d = yaml2dict(os.path.join(path, file_name))
        elif suffix == '.toml':
            d = toml.load(os.path.join(path, file_name))
        element = d["element"]
        for el in d["element"]:
            if el["style"] == "tabpanel":
                # Loop through tabs
                for tab in el["tab"]:
                    if "element" in tab:
                        if type(tab["element"]) is str:
                            # Must be a file
                            tab["element"] = self.read_gui_elements(path, tab["element"])
                    else:
                        tab["element"] = []
        return element

    # def find_id_in_config(self, id, config):
    #     # Find element in config file
    #     # First look in the menu
    #     for el in self.config["element"]:
    #         if el["id"] == id:
    #             return el
    #         if el["style"] == "tabpanel":
    #             for tab in el["tab"]:
    #                 if "element" in tab:
    #                     for el2 in tab["element"]:
    #                         if el2["id"] == id:
    #                             return el2
    #     return None

    # def font_metrics(self):
    #     font = self.qtapp.font()
    #     if self.framework == "pyqt5":
    #         from PyQt5.QtGui import QFontMetrics
    #         qfm = QFontMetrics(font)    
    #     elif self.framework == "pyside6":
    #         from PySide6.QtGui import QFontMetrics
    #         qfm = QFontMetrics(font)    
    #     return QFontMetrics(font)

    def string_width(self, string):
        font = self.qtapp.font()
        if self.framework == "pyqt5":
            from PyQt5.QtGui import QFontMetrics
            wdt = QFontMetrics(font).width(string)  
        elif self.framework == "pyside6":
            from PySide6.QtGui import QFontMetrics
            wdt = QFontMetrics(font).horizontalAdvance(string)
        return wdt

    def quit(self):
        self.qtapp.quit()

def yaml2dict(file_name):
    file = open(file_name,"r")
    dct = yaml.load(file, Loader=yaml.FullLoader)
    return dct



import os
from PySide6 import QtWebEngineWidgets
from PySide6 import QtCore, QtWidgets, QtWebChannel
from PySide6 import QtWebEngineCore
import json
from urllib.request import urlopen
import urllib
from geopandas import GeoDataFrame

from guitares.map.layer import Layer, list_layers, find_layer_by_id


class WebEnginePage(QtWebEngineCore.QWebEnginePage):
    def __init__(self, view, print_messages):
        super().__init__(view)
        self.print_messages = print_messages

    def javaScriptConsoleMessage(self, level, message, lineNumber, sourceID):
        if self.print_messages:
            print("javaScriptConsoleMessage: ", level, message, lineNumber, sourceID)


class MapBoxCompare(QtWidgets.QWidget):
    def __init__(self, element):
        super().__init__(element.parent.widget)

        self.available_map_styles = []
        self.available_map_styles.append({"id": "streets-v12", "name": "Streets"})
        self.available_map_styles.append({"id": "light-v11", "name": "Light"})
        self.available_map_styles.append({"id": "dark-v11", "name": "Dark"})
        self.available_map_styles.append({"id": "satellite-v9", "name": "Satellite"})
        self.available_map_styles.append({"id": "satellite-streets-v12", "name": "Satellite Streets"})

        self.gui = element.gui
        self.element = element

        # Check if element_map_style is in available_map_styles
        if element.map_style not in [style["id"] for style in self.available_map_styles]:
            # If not, set to default
            element.map_style = "light-v11"
        # Check if element.map_style starts with "mapbox://styles/"
        if not element.map_style.startswith("mapbox://styles/"):
            # If not, add it
            element.map_style = "mapbox://styles/mapbox/" + element.map_style     

        file_name = os.path.join(self.gui.server_path, "js", "mapbox_compare_defaults.js")
        with open(file_name, "w") as f:
            f.write("var default_compare_style = '" + element.map_style + "';\n")
            f.write("var default_compare_center = [" + str(element.map_center[0]) + "," + str(element.map_center[1]) + "]\n")
            f.write("var default_compare_zoom = " + str(element.map_zoom) + ";\n")
            f.write("var default_compare_projection = '" + element.map_projection + "';\n")

        url = "http://localhost:" + str(self.gui.server_port) + "/mapbox_compare.html"
        self.url = url

        self.ready = False
        self.ready_a = False
        self.ready_b = False

        self.server_path = self.gui.server_path

        self.setGeometry(
            0, 0, -1, -1
        )  # this is necessary because otherwise an invisible widget sits over the top left hand side of the screen and block the menu

        view = self.view = QtWebEngineWidgets.QWebEngineView(element.parent.widget)
        channel = self.channel = QtWebChannel.QWebChannel()
        view.page().profile().clearHttpCache()

        self.set_geometry()

        page = WebEnginePage(view, self.gui.js_messages)
        view.setPage(page)
        view.page().setWebChannel(channel)

        channel.registerObject("MapBoxCompare", self)

        print("URL: " + url)

        view.load(QtCore.QUrl(url))

        self.callback_module = element.module

        self.layer = {}
        self.map_extent = None
        self.map_moved = None
        self.point_clicked_callback = None
        self.zoom = None

        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.reload)
        self.timer.setSingleShot(True)
        self.timer.start(1000)

    def reload(self):
        print("Reloading ...")
        self.view.page().setWebChannel(self.channel)
        self.channel.registerObject("MapBoxCompare", self)
        self.view.load(QtCore.QUrl(self.url))

    def set(self):
        pass

    def set_geometry(self):
        resize_factor = self.element.gui.resize_factor
        x0, y0, wdt, hgt = self.element.get_position()
        self.view.setGeometry(x0, y0, wdt, hgt)

    def take_screenshot(self, output_file):
        self.view.grab().save(output_file, b"PNG")

    @QtCore.Slot(str)
    def mapReady(self, inpstr):
        inp = json.loads(inpstr)
        coords = inp[0:2]
        self.map_extent = coords
        side = inp[2]
        if side == "a":
            self.ready_a = True
        elif side == "b":
            self.ready_b = True
        if self.ready_a and self.ready_b:
            self.ready = True
        if self.ready:
            if hasattr(self.callback_module, "map_ready"):
                self.callback_module.map_ready()

    @QtCore.Slot(str)
    def layerStyleSet(self, coords):
        self.redraw_layers()

    # @QtCore.Slot(str)
    # def mouseMoved(self, coords):
    #     coords = json.loads(coords)
    #     self.map_extent = coords
    #     # Loop through layers to update each
    #     layers = list_layers(self.layer)
    #     for layer in layers:
    #         layer.update()
    #     if hasattr(self.callback_module, "map_moved"):
    #         self.callback_module.map_moved(coords)

    @QtCore.Slot(str)
    def mapMoved(self, coords):
        coords = json.loads(coords)
        self.map_extent = coords[0:2]
        self.map_center = coords[2:5]
        self.zoom = coords[4]
        # Loop through layers to update each
        layers = list_layers(self.layer)
        for layer in layers:
            layer.update()
        if hasattr(self.callback_module, "map_moved"):
            self.callback_module.map_moved(coords)

    # @QtCore.Slot(str)
    # def pointClicked(self, coords):
    #     coords = json.loads(coords)
    #     if self.point_clicked_callback:
    #         self.point_clicked_callback(coords)

    # @QtCore.Slot(str)
    # def getMapExtent(self, coords):
    #     coords = json.loads(coords)
    #     self.map_extent = coords

    # @QtCore.Slot(str, str)
    # def featureClicked(self, layer_id, feature_props):
    #     # Find layer by ID
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     if hasattr(layer, "select"):
    #         if layer.select:
    #             layer.select(json.loads(feature_props))

    # @QtCore.Slot(str, str, str)
    # def featureDrawn(self, feature_collection, feature_id, layer_id):
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     layer.feature_drawn(json.loads(feature_collection), feature_id)

    # @QtCore.Slot(str, str, str)
    # def featureModified(self, feature_collection, feature_id, layer_id):
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     layer.feature_modified(json.loads(feature_collection), feature_id)

    # @QtCore.Slot(str, str, str)
    # def featureSelected(self, feature_collection, feature_id, layer_id):
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     layer.feature_selected(json.loads(feature_collection), feature_id)

    # @QtCore.Slot(str)
    # def featureDeselected(self, layer_id):
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     if layer:
    #         layer.feature_deselected()

    # @QtCore.Slot(str, str, str)
    # def featureAdded(self, feature_collection, feature_id, layer_id):
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     layer.feature_added(json.loads(feature_collection), feature_id)

    # def get_extent(self):
    #     js_string = "import('/js/main.js').then(module => {module.getExtent()});"
    #     self.view.page().runJavaScript(js_string)

    # def click_point(self, callback):
    #     self.point_clicked_callback = callback
    #     self.runjs("/js/main.js", "clickPoint")

    # def set_center(self, lon, lat):
    #     self.runjs("/js/main.js", "setCenter", arglist=[lon, lat])

    # def set_zoom(self, zoom):
    #     self.runjs("/js/main.js", "setZoom", arglist=[zoom])

    # def fit_bounds(self, lon1, lat1, lon2, lat2):
    #     self.runjs("/js/main.js", "fitBounds", arglist=[lon1, lat1, lon2, lat2])

    def jump_to(self, lon, lat, zoom):
        self.runjs("/js/compare.js", "jumpTo", arglist=[lon, lat, zoom])

    def fly_to(self, lon, lat, zoom):
        self.runjs("/js/compare.js", "flyTo", arglist=[lon, lat, zoom])

    # def set_projection(self, projection):
    #     self.runjs("/js/main.js", "setProjection", arglist=[projection])

    def set_layer_style(self, style):
        self.runjs("/js/compare.js", "setLayerStyle", arglist=[style])

    def set_slider(self, npix):
        self.runjs("/js/compare.js", "setSlider", arglist=[npix])

    # def set_terrain(self, true_or_false, exaggeration):
    #     self.runjs("/js/main.js", "setTerrain", arglist=[true_or_false, exaggeration])

    # def set_mouse_default(self):
    #     self.runjs("/js/draw.js", "setMouseDefault", arglist=[])

    def add_layer(self, layer_id):
        # Adds a container layer
        if layer_id not in self.layer:
            self.layer[layer_id] = Layer(self, layer_id, layer_id)
            self.layer[layer_id].map_id = layer_id
        else:
            print("Layer " + layer_id + " already exists.")
        return self.layer[layer_id]

    def list_layers(self):
        # Return a list with all layers
        return list_layers(self.layer)

    def redraw_layers(self):
        # Redraw all layers (after map style has changed)
        layers = self.list_layers()
        for layer in layers:
            layer.redraw()

    def runjs(self, module, function, arglist=None):
        if not arglist:
            arglist = []
        string = "import('" + module + "').then(module => {module." + function + "("
        for iarg, arg in enumerate(arglist):
            if isinstance(arg, bool):
                if arg:
                    string = string + "true"
                else:
                    string = string + "false"
            elif isinstance(arg, int):
                string = string + str(arg)
            elif isinstance(arg, float):
                string = string + str(arg)
            elif isinstance(arg, dict):
                string = string + json.dumps(arg).replace('"',"'")
            elif isinstance(arg, list):
                string = string + json.dumps(arg).replace('"',"'")
            elif isinstance(arg, GeoDataFrame):
                if len(arg) == 0:
                    string = string + "{}"
                else:
                    string = string + arg.to_json()
            else:
                string = string + "'" + arg + "'"
            if iarg < len(arglist) - 1:
                string = string + ","
        string = string + ")});"
        self.view.page().runJavaScript(string)



from PySide6.QtWidgets import QTableView
from PySide6.QtCore import QSortFilterProxyModel
from PySide6.QtWidgets import QLabel
from PySide6 import QtCore
import traceback

from enum import Enum

import pandas as pd
import copy

class DataFrameModel(QtCore.QAbstractTableModel):
    """Class to create table view from dataframe"""

    DtypeRole = QtCore.Qt.UserRole + 1000
    ValueRole = QtCore.Qt.UserRole + 1001

    def __init__(self, df=pd.DataFrame(), parent=None):
        super(DataFrameModel, self).__init__(parent)
        self._dataframe = df

    def setDataFrame(self, dataframe):
        self.beginResetModel()
        self._dataframe = dataframe.copy()
        self.endResetModel()

    def dataFrame(self):
        return self._dataframe

    dataFrame = QtCore.pyqtProperty(pd.DataFrame, fget=dataFrame, fset=setDataFrame)

    @QtCore.pyqtSlot(int, QtCore.Qt.Orientation, result=str)
    def headerData(
        self,
        section: int,
        orientation: QtCore.Qt.Orientation,
        role: int = QtCore.Qt.DisplayRole,
    ):
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                return self._dataframe.columns[section]
            else:
                return str(self._dataframe.index[section])
        return QtCore.QVariant()

    def rowCount(self, parent=QtCore.QModelIndex()):
        if parent.isValid():
            return 0
        return len(self._dataframe.index)

    def columnCount(self, parent=QtCore.QModelIndex()):
        if parent.isValid():
            return 0
        return self._dataframe.columns.size

    def data(self, index, role):
        if not index.isValid() or not (
            0 <= index.row() < self.rowCount()
            and 0 <= index.column() < self.columnCount()
        ):
            return QtCore.QVariant()
        row = self._dataframe.index[index.row()]
        col = self._dataframe.columns[index.column()]
        dt = self._dataframe[col].dtype

        val = self._dataframe.iloc[row][col]
        if role == QtCore.Qt.DisplayRole:
            return str(val)
        elif role == DataFrameModel.ValueRole:
            return val
        if role == DataFrameModel.DtypeRole:
            return dt
        return QtCore.QVariant()

    def roleNames(self):
        roles = { 
            QtCore.Qt.DisplayRole: b"display",
            DataFrameModel.DtypeRole: b"dtype",
            DataFrameModel.ValueRole: b"value",
        }
        return roles

    def sort(self, column, order):
        colname = self._dataframe.columns.tolist()[column]
        self.layoutAboutToBeChanged.emit()
        self._dataframe.sort_values(
            colname, ascending=order == QtCore.Qt.AscendingOrder, inplace=True
        )
        self._dataframe.reset_index(inplace=True, drop=True)
        self.layoutChanged.emit()


class SelectionTable(Enum):
    NoSelection = 0
    SingleSelection = 1
    MultiSelection = 2
    ExtendedSelection = 3


class TableView(QTableView):
    """Table pyqt Element"""

    def __init__(self, element):
        super().__init__(element.parent.widget)
        self.element = element

        # Add title
        if element.text:
            label = QLabel(element.text, element.parent.widget)
            label.setStyleSheet("background: transparent; border: none")
            self.text_widget = label

        # Connect callback when event happens
        self.clicked.connect(self.callback)

        # If the header is clicked, the table is sorted and header_clicker is called
        if self.element.sortable:
            self.horizontalHeader().sectionClicked.connect(self.header_clicked)
            self.setSortingEnabled(True)
            self.horizontalHeader().setSortIndicatorShown(True)

#        self.horizontalHeader().setStyleSheet('QHeaderView::down-arrow { subcontrol-position: center right}')
        # self.selectionModel().selectionChanged.connect(self.callback)
        # Make sure that the sorting direction is shown

        self.set_geometry()

        # This allows for whole row selection and not individual cells
        self.setSelectionBehavior(QTableView.SelectRows)

        # Set selection mode
        selection_type = None
        if hasattr(self.element, "selection_type"):
            selection_type = getattr(SelectionTable, element.selection_type, None)
        self.multi_selection = False
        if selection_type:
            self.setSelectionMode(selection_type.value)
            if (
                selection_type == SelectionTable.MultiSelection
                or selection_type == SelectionTable.ExtendedSelection
            ):
                self.multi_selection = True

        self.execute_callback = True

        self.sort_order = 0
        self.sort_column = 0

    def set(self):

        # option_value
        df = self.element.getvar(self.element.option_value.variable_group, self.element.option_value.variable)
        if df is not None:
            self.df = df # Original unsorted dataframe
            df_sorted = copy.copy(df.reset_index()) # Sorted dataframe
            df_sorted = df_sorted.drop("index", axis=1)

        # Update items
        self.clearSpans()
        model = DataFrameModel(df_sorted)
        self.setModel(model)
        self.resizeColumnsToContents()
        self.verticalHeader().setVisible(False)

        if self.element.sortable:
            self.sortByColumn(self.sort_column, self.sort_order)

        # In case no selection is used skip
        if not self.element.variable:
            return

        # Find indices in sorted dataframe
        indices = self.find_sorted_indices()
        self.select_rows(indices)
        self.execute_callback = True

    def callback(self):
        # Find selected indices in the original dataframe
        if not self.element.variable:
            return
        indices = self.find_original_indices()
        # Set the new value 
        name = self.element.variable
        group = self.element.variable_group
        self.element.setvar(group, name, indices)
        try:
            # Execute callback
            if self.isEnabled() and self.element.callback:
                self.element.callback(indices, self)
            # Update GUI
            self.element.window.update()
        except:
            traceback.print_exc()

    def header_clicked(self, column_index):
        """Callback when header is clicked to sort"""
        if not self.element.variable:
            return
        self.sort_column = column_index
        self.sort_order = self.horizontalHeader().sortIndicatorOrder()
        indices = self.find_sorted_indices()
        self.select_rows(indices)

    def select_rows(self, indices):
        # Select rows in table
        self.selectionModel().clearSelection()
        for i in indices:
            self.selectRow(i)

    def find_original_indices(self):
        # Find indices of row in original dataframe df0 that match the row in the sorted dataframe df
        indices = [index.row() for index in self.selectionModel().selectedRows()]
        df0 = self.df
        df = self.model()._dataframe
        for index in indices:
            row = df.iloc[index]
            index0 = df0.index[df0.apply(lambda r: r.equals(row), axis=1)].tolist()[0]
            indices[indices.index(index)] = index0
        return indices
        
    def find_sorted_indices(self):
        # Find indices of row in sorted dataframe df0 that match the row in the original dataframe df
        indices0 = self.element.getvar(self.element.variable_group, self.element.variable)
        df0 = self.df
        df = self.model()._dataframe
        indices = []
        for index in indices0:
            row = df0.iloc[index]
            index0 = df.index[df.apply(lambda r: r.equals(row), axis=1)].tolist()[0]
            indices.append(index0)
        return indices

    def set_geometry(self):
        resize_factor = self.element.gui.resize_factor
        x0, y0, wdt, hgt = self.element.get_position()
        self.setGeometry(x0, y0, wdt, hgt)
        if self.element.text:
            label = self.text_widget
            fm = label.fontMetrics()
            wlab = int(fm.size(0, self.element.text).width())
            if (
                self.element.text_position == "above-center"
                or self.element.text_position == "above"
            ):
                label.setAlignment(QtCore.Qt.AlignCenter)
                label.setGeometry(
                    x0, int(y0 - 20 * resize_factor), wdt, int(20 * resize_factor)
                )
            elif self.element.text_position == "above-left":
                label.setAlignment(QtCore.Qt.AlignLeft)
                label.setGeometry(
                    x0, int(y0 - 20 * resize_factor), wlab, int(20 * resize_factor)
                )
            else:
                # Assuming left
                label.setAlignment(QtCore.Qt.AlignRight)
                label.setGeometry(
                    int(x0 - wlab - 3 * resize_factor),
                    int(y0 + 5 * self.element.gui.resize_factor),
                    wlab,
                    int(20 * resize_factor),
                )



import importlib
import os
import pathlib

from guitares.element import Element
from guitares.menu import Menu
# from guitares.statusbar import StatusBar

from guitares.dialog import window_dialog

class MenuBar:
    def __init__(self):
        self.parent = None
        self.widget = None

class Window:
    def __init__(self, config_dict, gui, type="main", data=None):
        self.type   = type
        self.gui    = gui
        self.data   = data
        self.width  = 800
        self.height = 600
        self.minimum_width  = 400
        self.minimum_height = 300
        self.maximize = False
        self.fixed_size = False
        self.title  = ""
        self.module = None
        self.variable_group = "_main"
        self.module = None
        self.method = None
        self.icon   = None
        self.modal  = False
        self.okay_method = None
        if "width" in config_dict["window"]:
            self.width = config_dict["window"]["width"]
        if "height" in config_dict["window"]:
            self.height = config_dict["window"]["height"]
        if "minimum_width" in config_dict["window"]:
            self.minimum_width = config_dict["window"]["minimum_width"]
        else:
            self.minimum_width = self.width
        if "minimum_height" in config_dict["window"]:
            self.minimum_height = config_dict["window"]["minimum_height"]
        else:
            self.minimum_height = self.height
        if "maximize" in config_dict["window"]:
            self.maximize = config_dict["window"]["maximize"]
        if "fixed_size" in config_dict["window"]:
            self.fixed_size = config_dict["window"]["fixed_size"]
        if "title" in config_dict["window"]:
            self.title = config_dict["window"]["title"]
        if "module" in config_dict["window"]:
            self.module = importlib.import_module(config_dict["window"]["module"])
        if "method" in config_dict["window"]:
            self.method = config_dict["window"]["method"]
        if "variable_group" in config_dict["window"]:
            self.variable_group = config_dict["window"]["variable_group"]
        if "icon" in config_dict["window"]:
            self.module = config_dict["window"]["icon"]
        if "modal" in config_dict["window"]:
            self.modal = config_dict["window"]["modal"]
        if self.module and "okay_method" in config_dict["window"]:
            self.okay_method = getattr(self.module, config_dict["window"]["okay_method"])
        if self.fixed_size:
            self.maximize=False    
        self.elements = []
        self.menus    = []
        self.toolbar  = []
        self.statusbar = {}
        # self.statusbar_fields = []

        if self.type == "popup":
            # Add OK and Cancel elements
            cancel = {'style': 'pushbutton', 'text': 'Cancel',
                      'position': {'x': -100, 'y': 10, 'width': 50, 'height': 20}, 'method': self.cancel,
                      'dependency': [], 'enable': True}
            ok = {'style': 'pushbutton', 'text': 'OK', 'position': {'x': -30, 'y': 10, 'width': 50, 'height': 20},
                  'method': self.ok, 'dependency': [], 'enable': True}
            config_dict["element"].append(cancel)
            config_dict["element"].append(ok)

        self.add_elements_to_tree(config_dict["element"], self, self)
        self.add_menu_to_tree(self.menus, config_dict["menu"], self)

        self.statusbar_fields = [] # temporary list
        if "statusbar" in config_dict:
            if "field" in config_dict["statusbar"]:
                for field in config_dict["statusbar"]["field"]:
                    self.statusbar_fields.append(field)

    def add_elements_to_tree(self, dcts, parent, window):
        parent.elements = []
        for dct in dcts:
            element = Element(dct, parent, window)
            if element.style == "tabpanel":
                for itab, tab_dct in enumerate(dct["tab"]):
                    if "element" in tab_dct:
                        self.add_elements_to_tree(tab_dct["element"],
                                                  element.tabs[itab],
                                                  window)
            elif dct["style"] == "panel":
                if "element" in dct:
                    self.add_elements_to_tree(dct["element"],
                                              element,
                                              window)
            parent.elements.append(element)

    def add_menu_to_tree(self, menu_list, dcts, parent):
        for dct in dcts:
            menu = Menu(dct, parent)
            if "menu" in dct:
                self.add_menu_to_tree(menu.menus, dct["menu"], menu)
            menu_list.append(menu)

    def build(self):

        if self.gui.framework == "pyqt5":
            from PyQt5.QtWidgets import QWidget, QVBoxLayout, QStatusBar, QLabel, QFrame
        elif self.gui.framework == "pyside6":
            from PySide6.QtWidgets import QWidget, QVBoxLayout, QStatusBar, QLabel, QFrame

        if self.type=="main":

            # Add main window

            mod = importlib.import_module(f"guitares.{self.gui.framework}.main_window")    
            window = mod.MainWindow(self)

            # Add menu
            menu_bar = MenuBar()
            menu_bar.widget = window.menuBar()
            self.add_menus(self.menus, menu_bar, self.gui)

            # Add toolbar
            # TODO

            # Status bar
            # self.window.statusBar().showMessage('Message in statusbar.')


            # Central widget
            central_widget = QWidget()
            layout = QVBoxLayout()
            central_widget.setLayout(layout)
            window.setCentralWidget(central_widget)
            self.widget = central_widget
            self.window_widget = window

        else:

            # Add pop-up window
            mod = importlib.import_module(f"guitares.{self.gui.framework}.popup_window")
            window = mod.PopupWindow(self)
            self.widget = window

        # Add elements
        self.add_elements(self.elements)

        # Add status bar (unless it is an empty dict)
        if self.statusbar_fields:
            mod = importlib.import_module(f"guitares.{self.gui.framework}.statusbar")
            self.statusbar = mod.StatusBar(self)

        # Set elements
        self.update()

        if self.maximize:
            window.showMaximized()

        window.show()

        # Check if a call start up callback function is needed
        if self.module and self.method:
            start_up_fcn = getattr(self.module, self.method)
            start_up_fcn()
            # Set elements again
            self.update()

        self.resize()

        if self.type == "popup":
            window.exec_()

        return window

    def ok(self, *args):
        # If there is no okay method or the okay method returns True, close the window. Otherwise, do nothing.
        if not self.okay_method or self.okay_method():
            self.widget.done(1) 

    def cancel(self, *args):
        self.widget.done(0)

    def closeEvent(self, event):
        pass

    def update(self):
        # Update all elements and menus
        self.set_elements(self.elements)
        self.set_menus(self.menus)

    def resize(self):
        # Resize elements
        self.resize_elements(self.elements)

    def add_elements(self, elements):
        # Loop through elements list
        for element in elements:
            if element.style == "tabpanel":
                # Add tab panel
                element.add()
                # Loop through tabs in tab panel
                for tab in element.tabs:
                    # And now add the elements in this tab
                    if tab.elements:
                        self.add_elements(tab.elements)
            elif element.style == "panel":
                # Add panel
                element.add()
                # And now add the elements in this frame
                if element.elements:
                    self.add_elements(element.elements)
            else:
                element.add()

    def set_elements(self, elements):
        # Loop through elements list
        for element in elements:
            try:
                if element.visible:
                    if element.style == "tabpanel":
                        # Set possible tab dependencies
                        element.set_dependencies()
                        # Only update the elements in the active tab
                        index = element.widget.currentIndex()
                        # Loop through elements in tab
                        for j, tab in enumerate(element.tabs):
                            # Check if tab has elements
                            if tab.elements and j==index:
                                # And now add the elements in this tab
                                self.set_elements(tab.elements)
                    elif element.style == "panel":
                        # Set the dependencies
                        element.set_dependencies()
                        # Check if this frame has elements
                        if element.elements:
                            # Set elements in this frame
                            self.set_elements(element.elements)
                    else:
                        # Set the element values
                        element.widget.set()
                        # Set the dependencies
                        element.set_dependencies()
            except Exception as err:
                print(err)

    def find_element_by_id(self, element_id, elements=None):
        element_found = None
        if elements == None:
            elements=self.elements
        for element in elements:
            if element.id == element_id:
                return element
            if element.style == "tabpanel":
                # Loop through tabs
                for tab in element.tabs:
                    # Look for elements in this tab
                    if tab.elements:
                        element_found = self.find_element_by_id(element_id, elements=tab.elements)
                        if element_found:
                            return element_found
            elif element.style == "panel":
                # Look for elements in this frame
                if element.elements:
                    element_found = self.find_element_by_id(element_id, elements=element.elements)
                    if element_found:
                        return element_found
        return None

    def resize_elements(self, elements):
        # Loop through elements
        for element in elements:
            # Set geometry of this element
            element.set_geometry()
            # Check for children
            if element.style == "tabpanel":
                # Loop through tabs
                for tab in element.tabs:
                    # And resize elements in this tab
                    if tab.elements:
                        self.resize_elements(tab.elements)
            elif element.style == "panel":
                self.resize_elements(element.elements)

    def add_menus(self, menus, parent, gui):
        # Loop through elements list
        for menu in menus:
            menu.parent = parent
            menu.gui = gui
            menu.add()
            if menu.menus:
                self.add_menus(menu.menus, menu, gui)

    def add_menu_from_dict(self, menu_dict, parent_id, has_children=False):
        # Add a menu from a dictionary
        parent = self.find_menu_item_by_id(parent_id)
        menu = Menu(menu_dict, parent)
        parent.menus.append(menu)
        menu.add(has_children=has_children)

    def set_menus(self, menus):
        for menu in menus:
            if menu.menus:
                self.set_menus(menu.menus)
            else:
                menu.set_dependencies()

    def find_menu_item_by_id(self, menu_id, menus=None):
        if menus == None:
            menus = self.menus
        for menu in menus:
            if menu.id == menu_id:
                return menu
            item = self.find_menu_item_by_id(menu_id, menus=menu.menus)
            if item:
                return item
        return None


    def dialog_ok_cancel(self, text, title=" "):
        return window_dialog(self, text, type="question", title=title)

    def dialog_yes_no(self, text, title=" "):
        return window_dialog(self, text, type="question_yes_no", title=title)

    def dialog_custom(self, text, title=" ", button_text=None):
        return window_dialog(self, text, type="custom", title=title, button_text=button_text)

    def dialog_warning(self, text, title="Warning"):
        window_dialog(self, text, type="warning", title=title)

    def dialog_info(self, text, title=" "):
        window_dialog(self, text, type="info", title=title)

    def dialog_critical(self, text, title="Critical"):
        window_dialog(self, text, type="critical", title=title)

    def dialog_progress(self, text, nmax, title=" "):
        dlg = window_dialog(self, text, title=title, type="progress", nmax=nmax)
        return dlg

    def dialog_wait(self, text, title=" "):
        dlg = window_dialog(self, text, title=title, type="wait")
        return dlg

    def dialog_open_file(self, text, filter, path=None, file_name=None, selected_filter=None, allow_directory_change=True):
        if path == None:
            path = os.getcwd()
        results = window_dialog(self, text, type="open_file", path=path, file_name=file_name, filter=filter, selected_filter=selected_filter)
        full_name = results[0]
        if not full_name:
            return "", "", "", "", ""
        fltr = results[1]
        pth = pathlib.Path(full_name)
        if not allow_directory_change:
            if str(pth.parent) != path:
                self.dialog_warning("Sorry, you cannot select a file from another directory !")
                return "", "", "", "", ""
        path = str(pth.parent)
        name = pth.name
        ext = pth.suffix
        return full_name, path, name, ext, fltr

    def dialog_save_file(self, text, filter, path=None, file_name=None, selected_filter=None, allow_directory_change=True):
        if path == None:
            path = os.getcwd()
        results = window_dialog(self, text, type="save_file", path=path, file_name=file_name, filter=filter, selected_filter=selected_filter)
        full_name = results[0]
        if not full_name:
            return "", "", "", "", ""
        fltr = results[1]
        pth = pathlib.Path(full_name)
        if not allow_directory_change:
            if str(pth.parent) != path:
                self.dialog_warning("Sorry, you cannot save this file to another directory !")
                return "", "", "", "", ""
        path = str(pth.parent)
        name = pth.name
        ext = pth.suffix
        return full_name, path, name, ext, fltr

    def dialog_select_path(self, text, path=None):
        new_path = window_dialog(self, text, filter=filter, type="select_path", path=path)
        return new_path
 
    def dialog_string(self, text, title=" "):
        return window_dialog(self, text, type="string", title=title)



import os
from PySide6 import QtWebEngineWidgets, QtWebEngineCore, QtCore, QtWidgets, QtWebChannel
import json
# from urllib.request import urlopen
# import urllib
from geopandas import GeoDataFrame
from pandas import DataFrame

from guitares.map.layer import Layer, list_layers

class WebEnginePage(QtWebEngineCore.QWebEnginePage):
    def __init__(self, view, print_messages):
        super().__init__(view)
        self.print_messages = print_messages

    def javaScriptConsoleMessage(self, level, message, lineNumber, sourceID):
        if self.print_messages:
            print("javaScriptConsoleMessage: ", level, message, lineNumber, sourceID)


class MapLibreCompare(QtWidgets.QWidget):
    def __init__(self, element):
        super().__init__(element.parent.widget)

        self.gui = element.gui
        self.element = element

        self.callback_module = element.module

        file_name = os.path.join(self.gui.server_path, "js", "maplibre_compare_defaults.js")
        with open(file_name, "w") as f:
            f.write("var default_compare_style = '" + element.map_style + "';\n")
            f.write("var default_compare_center = [" + str(element.map_center[0]) + "," + str(element.map_center[1]) + "]\n")
            f.write("var default_compare_zoom = " + str(element.map_zoom) + ";\n")
            f.write("var default_compare_projection = '" + element.map_projection + "';\n")

        self.url = "http://localhost:" + str(self.gui.server_port)
        # url = "http://localhost:" + str(self.gui.server_port) + "/maplibre_compare.html"
        # self.url = url

        self.ready = False
        self.ready_a = False
        self.ready_b = False

        self.webchannel_ok = False
        # self.ready = False

        self.server_path = self.gui.server_path

        self.setGeometry(
            0, 0, -1, -1
        )  # this is necessary because otherwise an invisible widget sits over the top left hand side of the screen and block the menu

        self.view = QtWebEngineWidgets.QWebEngineView(element.parent.widget)
        self.view.setPage(WebEnginePage(self.view, self.gui.js_messages))
        self.view.page().settings().setAttribute(QtWebEngineCore.QWebEngineSettings.WebAttribute.WebGLEnabled, True)
        self.view.page().settings().setAttribute(QtWebEngineCore.QWebEngineSettings.LocalContentCanAccessRemoteUrls, True)

        self.channel = QtWebChannel.QWebChannel()
        self.channel.registerObject("MapLibreCompare", self)
        self.view.page().setWebChannel(self.channel)

        self.set_geometry()
        self.view.loadFinished.connect(self.load_finished)
        self.view.setUrl(QtCore.QUrl(self.url + "/maplibre_compare.html"))

        # self.server_path = self.gui.server_path

        # self.setGeometry(
        #     0, 0, -1, -1
        # )  # this is necessary because otherwise an invisible widget sits over the top left hand side of the screen and block the menu

        # view = self.view = QtWebEngineWidgets.QWebEngineView(element.parent.widget)
        # channel = self.channel = QtWebChannel.QWebChannel()
        # view.page().profile().clearHttpCache()

        # self.set_geometry()

        # page = WebEnginePage(view, self.gui.js_messages)
        # view.setPage(page)
        # view.page().setWebChannel(channel)

        # channel.registerObject("MapLibreCompare", self)

        # view.load(QtCore.QUrl(url))

        # self.callback_module = element.module

        self.layer = {}
        self.map_extent = None
        self.map_moved = None
        self.point_clicked_callback = None
        self.zoom = None

        # self.timer = QtCore.QTimer()
        # # self.timer.timeout.connect(self.reload)
        # self.timer.setSingleShot(True)
        # self.timer.start(1000)


    def load_finished(self, message):
        self.timer_ping = QtCore.QTimer()        
        self.timer_ping.timeout.connect(self.ping)
        self.timer_ping.start(1000)

    def ping(self):
        # Sending a ping to main.js
        print("Ping!")
        self.runjs("/js/compare.js", "ping", arglist=["ping"])

    @QtCore.Slot(str)
    def pong(self, message):
        # Python heard a pong!
        self.timer_ping.stop()
        print("Pong received! Adding map compare ...")
        # Add map
        self.runjs("/js/compare.js", "addMap", arglist=[])

    def set(self):
        pass

    def set_geometry(self):
        resize_factor = self.element.gui.resize_factor
        x0, y0, wdt, hgt = self.element.get_position()
        self.view.setGeometry(x0, y0, wdt, hgt)

    def take_screenshot(self, output_file):
        self.view.grab().save(output_file, b"PNG")

    @QtCore.Slot(str)
    def mapReady(self, inpstr):
        inp = json.loads(inpstr)
        coords = inp[0:2]
        self.map_extent = coords
        side = inp[2]
        if side == "a":
            self.ready_a = True
        elif side == "b":
            self.ready_b = True
        if self.ready_a and self.ready_b:
            self.ready = True
        if self.ready:
            if hasattr(self.callback_module, "map_ready"):
                self.callback_module.map_ready()

    @QtCore.Slot(str)
    def layerStyleSet(self, coords):
        self.redraw_layers()

    # @QtCore.Slot(str)
    # def mouseMoved(self, coords):
    #     coords = json.loads(coords)
    #     self.map_extent = coords
    #     # Loop through layers to update each
    #     layers = list_layers(self.layer)
    #     for layer in layers:
    #         layer.update()
    #     if hasattr(self.callback_module, "map_moved"):
    #         self.callback_module.map_moved(coords)

    @QtCore.Slot(str)
    def mapMoved(self, coords):
        coords = json.loads(coords)
        self.map_extent = coords[0:2]
        self.map_center = coords[2:5]
        self.zoom = coords[4]
        # Loop through layers to update each
        layers = list_layers(self.layer)
        for layer in layers:
            layer.update()
        if hasattr(self.callback_module, "map_moved"):
            self.callback_module.map_moved(coords)

    # @QtCore.Slot(str)
    # def pointClicked(self, coords):
    #     coords = json.loads(coords)
    #     if self.point_clicked_callback:
    #         self.point_clicked_callback(coords)

    # @QtCore.Slot(str)
    # def getMapExtent(self, coords):
    #     coords = json.loads(coords)
    #     self.map_extent = coords

    # @QtCore.Slot(str, str)
    # def featureClicked(self, layer_id, feature_props):
    #     # Find layer by ID
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     if hasattr(layer, "select"):
    #         if layer.select:
    #             layer.select(json.loads(feature_props))

    # @QtCore.Slot(str, str, str)
    # def featureDrawn(self, feature_collection, feature_id, layer_id):
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     layer.feature_drawn(json.loads(feature_collection), feature_id)

    # @QtCore.Slot(str, str, str)
    # def featureModified(self, feature_collection, feature_id, layer_id):
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     layer.feature_modified(json.loads(feature_collection), feature_id)

    # @QtCore.Slot(str, str, str)
    # def featureSelected(self, feature_collection, feature_id, layer_id):
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     layer.feature_selected(json.loads(feature_collection), feature_id)

    # @QtCore.Slot(str)
    # def featureDeselected(self, layer_id):
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     if layer:
    #         layer.feature_deselected()

    # @QtCore.Slot(str, str, str)
    # def featureAdded(self, feature_collection, feature_id, layer_id):
    #     layer = find_layer_by_id(layer_id, self.layer)
    #     layer.feature_added(json.loads(feature_collection), feature_id)

    # def get_extent(self):
    #     js_string = "import('/js/main.js').then(module => {module.getExtent()});"
    #     self.view.page().runJavaScript(js_string)

    # def click_point(self, callback):
    #     self.point_clicked_callback = callback
    #     self.runjs("/js/main.js", "clickPoint")

    # def set_center(self, lon, lat):
    #     self.runjs("/js/main.js", "setCenter", arglist=[lon, lat])

    # def set_zoom(self, zoom):
    #     self.runjs("/js/main.js", "setZoom", arglist=[zoom])

    # def fit_bounds(self, lon1, lat1, lon2, lat2):
    #     self.runjs("/js/main.js", "fitBounds", arglist=[lon1, lat1, lon2, lat2])

    def jump_to(self, lon, lat, zoom):
        self.runjs("/js/compare.js", "jumpTo", arglist=[lon, lat, zoom])

    def fly_to(self, lon, lat, zoom):
        self.runjs("/js/compare.js", "flyTo", arglist=[lon, lat, zoom])

    # def set_projection(self, projection):
    #     self.runjs("/js/main.js", "setProjection", arglist=[projection])

    def set_layer_style(self, style):
        self.runjs("/js/compare.js", "setLayerStyle", arglist=[style])

    def set_slider(self, npix):
        self.runjs("/js/compare.js", "setSlider", arglist=[npix])

    # def set_terrain(self, true_or_false, exaggeration):
    #     self.runjs("/js/main.js", "setTerrain", arglist=[true_or_false, exaggeration])

    # def set_mouse_default(self):
    #     self.runjs("/js/draw.js", "setMouseDefault", arglist=[])

    def add_layer(self, layer_id):
        # Adds a container layer
        if layer_id not in self.layer:
            self.layer[layer_id] = Layer(self, layer_id, layer_id)
            self.layer[layer_id].map_id = layer_id
        else:
            print("Layer " + layer_id + " already exists.")
        return self.layer[layer_id]

    def list_layers(self):
        # Return a list with all layers
        return list_layers(self.layer)

    def redraw_layers(self):
        # Redraw all layers (after map style has changed)
        layers = self.list_layers()
        for layer in layers:
            layer.redraw()


    def runjs(self, module, function, arglist=None):
        if not arglist:
            arglist = []
        string = "import('" + self.url + module + "').then(module => {module." + function + "("
        for iarg, arg in enumerate(arglist):
            if isinstance(arg, bool):
                if arg:
                    string = string + "true"
                else:
                    string = string + "false"
            elif isinstance(arg, int):
                string = string + str(arg)
            elif isinstance(arg, float):
                string = string + str(arg)
            elif isinstance(arg, dict):
                string = string + json.dumps(arg).replace('"',"'")
            elif isinstance(arg, list):
                string = string + json.dumps(arg).replace('"',"'")
            elif isinstance(arg, tuple):
                string = string + json.dumps(arg).replace('"',"'")
            elif isinstance(arg, GeoDataFrame):
                if len(arg) == 0:
                    string = string + "{}"
                else:
                    # Need to remove timeseries from geodataframe
                    for columnName, columnData in arg.items():
                        if isinstance(columnData.iloc[0], DataFrame):
                            arg = arg.drop([columnName], axis=1)
                    string = string + arg.to_json()
            elif arg is None:
                string = string + "null"        
            else:
                string = string + "'" + arg + "'"
            if iarg < len(arglist) - 1:
                string = string + ","
        string = string + ")});"
        # print(string)
        self.view.page().runJavaScript(string)



    # def runjs(self, module, function, arglist=None):
    #     if not arglist:
    #         arglist = []
    #     string = "import('" + module + "').then(module => {module." + function + "("
    #     for iarg, arg in enumerate(arglist):
    #         if isinstance(arg, bool):
    #             if arg:
    #                 string = string + "true"
    #             else:
    #                 string = string + "false"
    #         elif isinstance(arg, int):
    #             string = string + str(arg)
    #         elif isinstance(arg, float):
    #             string = string + str(arg)
    #         elif isinstance(arg, dict):
    #             string = string + json.dumps(arg).replace('"',"'")
    #         elif isinstance(arg, list):
    #             string = string + json.dumps(arg).replace('"',"'")
    #         elif isinstance(arg, GeoDataFrame):
    #             if len(arg) == 0:
    #                 string = string + "{}"
    #             else:
    #                 string = string + arg.to_json()
    #         else:
    #             string = string + "'" + arg + "'"
    #         if iarg < len(arglist) - 1:
    #             string = string + ","
    #     string = string + ")});"
    #     self.view.page().runJavaScript(string)
